# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

from __future__ import annotations

import enum
import json
import numpy as np
import typing
import torch

from typing import Dict

from embodichain.lab.sim.robots.dexforce_w1.types import (
    DexforceW1HandBrand,
    DexforceW1ArmSide,
    DexforceW1ArmKind,
    DexforceW1Version,
)
from embodichain.lab.sim.robots.dexforce_w1.utils import (
    build_dexforce_w1_cfg,
)
from embodichain.lab.sim.solvers import SolverCfg
from embodichain.lab.sim.cfg import (
    RobotCfg,
    JointDrivePropertiesCfg,
    RigidBodyAttributesCfg,
)
from embodichain.lab.sim.utility.cfg_utils import merge_robot_cfg
from embodichain.data import get_data_path
from embodichain.utils import configclass, logger


@configclass
class DexforceW1Cfg(RobotCfg):
    """DexforceW1 specific configuration, inherits from RobotCfg and allows custom parameters."""

    version: DexforceW1Version = DexforceW1Version.V021
    arm_kind: DexforceW1ArmKind = DexforceW1ArmKind.INDUSTRIAL
    with_default_eef: bool = True

    @classmethod
    def from_dict(
        cls, init_dict: Dict[str, str | float | tuple | dict]
    ) -> DexforceW1Cfg:
        """Initialize DexforceW1Cfg from a dictionary.

        Args:
            init_dict (Dict[str, str | float | tuple | dict): Dictionary of configuration parameters.

        Returns:
            DexforceW1Cfg: An instance of DexforceW1Cfg with parameters set.
        """

        init_dict_m = init_dict.copy()
        version = init_dict_m.get("version", "v021")
        arm_kind = init_dict_m.get("arm_kind", "anthropomorphic")
        with_default_eef = init_dict_m.get("with_default_eef", True)
        init_dict_m.pop("version", None)
        init_dict_m.pop("arm_kind", None)
        init_dict_m.pop("with_default_eef", None)

        cfg: DexforceW1Cfg = cls()._build_default_cfg(
            version=version, arm_kind=arm_kind, with_default_eef=with_default_eef
        )

        default_physics_cfgs = cls()._build_default_physics_cfgs(
            arm_kind=arm_kind, with_default_eef=with_default_eef
        )
        for key, value in default_physics_cfgs.items():
            setattr(cfg, key, value)

        default_solver_cfg = cls()._build_default_solver_cfg(
            is_industrial=(arm_kind == "industrial")
        )
        cfg.solver_cfg = default_solver_cfg

        cfg = merge_robot_cfg(cfg, init_dict_m)

        return cfg

    @staticmethod
    def _build_default_solver_cfg(is_industrial: bool) -> SolverCfg:
        from embodichain.lab.sim.solvers import SRSSolverCfg
        from embodichain.lab.sim.robots.dexforce_w1.params import (
            W1ArmKineParams,
        )

        if is_industrial:
            w1_left_arm_params = W1ArmKineParams(
                arm_side=DexforceW1ArmSide.LEFT,
                arm_kind=DexforceW1ArmKind.INDUSTRIAL,
                version=DexforceW1Version.V021,
            )
            w1_right_arm_params = W1ArmKineParams(
                arm_side=DexforceW1ArmSide.RIGHT,
                arm_kind=DexforceW1ArmKind.INDUSTRIAL,
                version=DexforceW1Version.V021,
            )
            left_arm_tcp = np.array(
                [
                    [1.0, 0.0, 0.0, 0.0],
                    [0.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 0.15],
                    [0.0, 0.0, 0.0, 1.0],
                ]
            )
            right_arm_tcp = np.array(
                [
                    [1.0, 0.0, 0.0, 0.0],
                    [0.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 0.15],
                    [0.0, 0.0, 0.0, 1.0],
                ]
            )
        else:
            w1_left_arm_params = W1ArmKineParams(
                arm_side=DexforceW1ArmSide.LEFT,
                arm_kind=DexforceW1ArmKind.ANTHROPOMORPHIC,
                version=DexforceW1Version.V021,
            )
            w1_right_arm_params = W1ArmKineParams(
                arm_side=DexforceW1ArmSide.RIGHT,
                arm_kind=DexforceW1ArmKind.ANTHROPOMORPHIC,
                version=DexforceW1Version.V021,
            )
            left_arm_tcp = np.array(
                [
                    [-1.0, 0.0, 0.0, 0.012],
                    [0.0, 0.0, 1.0, 0.0675],
                    [0.0, 1.0, 0.0, 0.127],
                    [0.0, 0.0, 0.0, 1.0],
                ]
            )
            right_arm_tcp = np.array(
                [
                    [1.0, 0.0, 0.0, 0.012],
                    [0.0, 0.0, -1.0, -0.0675],
                    [0.0, 1.0, 0.0, 0.127],
                    [0.0, 0.0, 0.0, 1.0],
                ]
            )

        return {
            "right_arm": SRSSolverCfg(
                end_link_name="right_ee",
                root_link_name="right_arm_base",
                dh_params=w1_right_arm_params.dh_params,
                qpos_limits=w1_right_arm_params.qpos_limits,
                T_e_oe=w1_right_arm_params.T_e_oe,
                T_b_ob=w1_right_arm_params.T_b_ob,
                link_lengths=w1_right_arm_params.link_lengths,
                rotation_directions=w1_right_arm_params.rotation_directions,
                tcp=right_arm_tcp,
            ),
            "left_arm": SRSSolverCfg(
                end_link_name="left_ee",
                root_link_name="left_arm_base",
                dh_params=w1_left_arm_params.dh_params,
                qpos_limits=w1_left_arm_params.qpos_limits,
                T_e_oe=w1_left_arm_params.T_e_oe,
                T_b_ob=w1_left_arm_params.T_b_ob,
                link_lengths=w1_left_arm_params.link_lengths,
                rotation_directions=w1_left_arm_params.rotation_directions,
                tcp=left_arm_tcp,
            ),
        }

    @staticmethod
    def _build_default_physics_cfgs(
        arm_kind: str, with_default_eef: bool = True
    ) -> typing.Dict[str, typing.Any]:
        """Build default physics configurations for DexforceW1.

        Args:
            arm_kind: Type of arm, either "industrial" or "anthropomorphic"
            with_default_eef: Whether to include default end-effector configurations

        Returns:
            Dictionary containing physics configuration parameters
        """
        # Define default joint drive parameters and corresponding joint name patterns
        DEFAULT_EEF_JOINT_DRIVE_PARAMS = {
            "stiffness": 1e2,
            "damping": 1e1,
            "max_effort": 1e3,
        }

        DEFAULT_EEF_HAND_JOINT_NAMES = (
            "(LEFT|RIGHT)_HAND_(THUMB[12]|INDEX|MIDDLE|RING|PINKY)"
        )

        DEFAULT_EEF_GRIPPER_JOINT_NAMES = "(LEFT|RIGHT)_FINGER[1-2]"

        # Define common joint patterns
        ARM_JOINTS = "(RIGHT|LEFT)_J[0-9]"
        BODY_JOINTS = "(ANKLE|KNEE|BUTTOCK|WAIST)"

        # Define physics parameters for different joint types
        joint_params = {
            "stiffness": {
                ARM_JOINTS: 1e4,
                BODY_JOINTS: 1e7,
            },
            "damping": {
                ARM_JOINTS: 1e3,
                BODY_JOINTS: 1e4,
            },
            "max_effort": {
                ARM_JOINTS: 1e5,
                BODY_JOINTS: 1e10,
            },
        }

        drive_pros = JointDrivePropertiesCfg(**joint_params)

        if with_default_eef:
            eef_joint_names = (
                DEFAULT_EEF_HAND_JOINT_NAMES
                if arm_kind == "anthropomorphic"
                else DEFAULT_EEF_GRIPPER_JOINT_NAMES
            )
            drive_pros.stiffness.update(
                {eef_joint_names: DEFAULT_EEF_JOINT_DRIVE_PARAMS["stiffness"]}
            )
            drive_pros.damping.update(
                {eef_joint_names: DEFAULT_EEF_JOINT_DRIVE_PARAMS["damping"]}
            )
            drive_pros.max_effort.update(
                {eef_joint_names: DEFAULT_EEF_JOINT_DRIVE_PARAMS["max_effort"]}
            )

        return {
            "min_position_iters": 32,
            "min_velocity_iters": 8,
            "drive_pros": drive_pros,
            "attrs": RigidBodyAttributesCfg(
                mass=1.0,
                static_friction=0.95,
                dynamic_friction=0.9,
                linear_damping=0.7,
                angular_damping=0.7,
                contact_offset=0.005,
                rest_offset=0.001,
                restitution=0.05,
                max_depenetration_velocity=10.0,
            ),
        }

    @staticmethod
    def _build_default_cfg(
        version: str = "v021",
        arm_kind: str = "anthropomorphic",
        with_default_eef: bool = True,
    ) -> DexforceW1Cfg:

        if arm_kind == "industrial":
            hand_types = {
                DexforceW1ArmSide.LEFT: DexforceW1HandBrand.DH_PGC_GRIPPER_M,
                DexforceW1ArmSide.RIGHT: DexforceW1HandBrand.DH_PGC_GRIPPER_M,
            }
        else:
            hand_types = {
                DexforceW1ArmSide.LEFT: DexforceW1HandBrand.BRAINCO_HAND,
                DexforceW1ArmSide.RIGHT: DexforceW1HandBrand.BRAINCO_HAND,
            }

        hand_versions = {
            DexforceW1ArmSide.LEFT: DexforceW1Version(version),
            DexforceW1ArmSide.RIGHT: DexforceW1Version(version),
        }

        cfg = build_dexforce_w1_cfg(
            arm_kind=DexforceW1ArmKind(arm_kind),
            hand_types=hand_types,
            hand_versions=hand_versions,
            include_hand=with_default_eef,
        )
        cfg.version = DexforceW1Version(version)
        cfg.arm_kind = DexforceW1ArmKind(arm_kind)
        cfg.with_default_eef = with_default_eef

        return cfg

    def to_dict(self):
        """Convert config to a Python dict, handling enums and numpy arrays."""

        def serialize(obj, _visited=None):
            if _visited is None:
                _visited = set()
            # Only skip recursion for mutable objects (dict, custom class)
            if isinstance(obj, (dict, object)) and not isinstance(
                obj, (str, int, float, bool, type(None))
            ):
                obj_id = id(obj)
                if obj_id in _visited:
                    return None  # Prevent infinite recursion
                _visited.add(obj_id)

            if isinstance(obj, enum.Enum):
                return obj.value
            if isinstance(obj, np.ndarray):
                return obj.tolist()
            if isinstance(obj, dict):
                # Only serialize values, keep keys as str/int/float/bool/None
                return {str(k): serialize(v, _visited) for k, v in obj.items()}
            if isinstance(obj, (list, tuple)):
                return [serialize(v, _visited) for v in obj]
            if hasattr(obj, "to_dict") and obj is not self:
                return serialize(obj.to_dict(), _visited)
            if hasattr(obj, "__dict__"):
                return {k: serialize(v, _visited) for k, v in obj.__dict__.items()}
            return obj

        return serialize(self)

    def to_string(self):
        """Return config as a JSON string."""
        return json.dumps(self.to_dict(), indent=2)

    def save_to_file(self, filepath):
        """Save config to a local file as JSON."""
        with open(filepath, "w") as f:
            f.write(self.to_string())

    def build_pk_serial_chain(
        self, device: torch.device = torch.device("cpu"), **kwargs
    ) -> Dict[str, "pk.SerialChain"]:
        from embodichain.lab.sim.utility.solver_utils import (
            create_pk_chain,
            create_pk_serial_chain,
        )

        if DexforceW1ArmKind.INDUSTRIAL == self.arm_kind:
            urdf_path = get_data_path("DexforceW1V021/DexforceW1_v02_2.urdf")
        elif DexforceW1ArmKind.ANTHROPOMORPHIC == self.arm_kind:
            urdf_path = get_data_path("DexforceW1V021/DexforceW1_v02_1.urdf")

        chain = create_pk_chain(urdf_path, device)

        left_arm_chain = create_pk_serial_chain(
            chain=chain, end_link_name="left_ee", root_link_name="left_arm_base"
        ).to(device=device)
        right_arm_chain = create_pk_serial_chain(
            chain=chain, end_link_name="right_ee", root_link_name="right_arm_base"
        ).to(device=device)

        return {
            "left_arm": left_arm_chain,
            "right_arm": right_arm_chain,
        }


if __name__ == "__main__":
    # Example usage
    import numpy as np

    np.set_printoptions(precision=5, suppress=True)
    from embodichain.lab.sim import SimulationManager, SimulationManagerCfg
    from embodichain.lab.sim.robots.dexforce_w1.types import (
        DexforceW1ArmKind,
    )

    config = SimulationManagerCfg(headless=True, sim_device="cpu", num_envs=4)
    sim = SimulationManager(config)

    cfg = DexforceW1Cfg.from_dict(
        {"uid": "dexforce_w1", "version": "v021", "arm_kind": "anthropomorphic"}
    )

    robot = sim.add_robot(cfg=cfg)
    sim.update(step=1)
    print("DexforceW1 robot added to the simulation.")

    from IPython import embed

    embed()



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

import os

embodichain_dir = os.path.dirname(__file__)


# Read version from VERSION file
def _get_version():
    version_file = os.path.join(embodichain_dir, "VERSION")
    try:
        with open(version_file, "r") as f:
            return f.read().strip()
    except FileNotFoundError:
        print("VERSION file not found.")
        return "unknown"


__version__ = _get_version()



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

from __future__ import annotations

import torch
import os
import random
import numpy as np
from typing import TYPE_CHECKING, Literal, Union, List

from dexsim.utility import images_to_video
from embodichain.lab.gym.envs.managers import Functor, FunctorCfg
from embodichain.lab.sim.sensors.camera import CameraCfg, Camera

if TYPE_CHECKING:
    from embodichain.lab.gym.envs import EmbodiedEnv


class record_camera_data(Functor):
    """Record camera data in the environment. The camera is usually setup with third-person view, and
    is used to record the scene during the episode. It is helpful for debugging and visualization.

    Note:
        Currently, the functor is implemented in `interval' mode such that, it can only save the
        recorded frames when in :meth:`env.step()` function call. For example:
        ```python
        env.step()
        # perform multiple steps in the same episode
        env.reset()
        env.step()  # the video of the first episode will be saved here.
        ```
        The final episode frames will not be saved in the current implementation.
        We may improve it in the future.
    """

    def __init__(self, cfg: FunctorCfg, env: EmbodiedEnv):
        """Initialize the functor.

        Args:
            cfg: The configuration of the functor.
            env: The environment instance.

        Raises:
            ValueError: If the asset is not a RigidObject or an Articulation.
        """
        super().__init__(cfg, env)

        # extract the used quantities (to enable type-hinting)
        self._name = cfg.params.get("name", "default")
        resolution = cfg.params.get("resolution", (640, 480))
        eye = cfg.params.get("eye", (0, 0, 2))
        target = cfg.params.get("target", (0, 0, 0))
        up = cfg.params.get("up", (0, 0, 1))
        intrinsics = cfg.params.get(
            "intrinsics", (600, 600, int(resolution[0] / 2), int(resolution[1] / 2))
        )

        self.camera: Camera = env.sim.add_sensor(
            sensor_cfg=CameraCfg(
                uid=self._name,
                width=resolution[0],
                height=resolution[1],
                extrinsics=CameraCfg.ExtrinsicsCfg(eye=eye, target=target, up=up),
                intrinsics=intrinsics,
            )
        )

        # Add this camera's group ID to the environment for batch rendering when RT is enabled.
        if getattr(env.sim, "is_rt_enabled", False):
            env.add_camera_group_id(self.camera.group_id)

        self._current_episode = 0
        self._frames: List[np.ndarray] = []

    def _draw_frames_into_one_image(self, frames: torch.Tensor) -> torch.Tensor:
        """
        Concatenate multiple frames into a single image with nearly square arrangement.

        Args:
            frames: Tensor with shape (B, H, W, 4) where B is batch size

        Returns:
            Single concatenated image tensor with shape (grid_h * H, grid_w * W, 4)
        """
        if frames.numel() == 0:
            return frames

        B, H, W, C = frames.shape

        # Calculate grid dimensions for nearly square arrangement
        grid_w = int(torch.ceil(torch.sqrt(torch.tensor(B, dtype=torch.float32))))
        grid_h = int(torch.ceil(torch.tensor(B, dtype=torch.float32) / grid_w))

        # Create empty grid to hold all frames
        result = torch.zeros(
            (grid_h * H, grid_w * W, C), dtype=frames.dtype, device=frames.device
        )

        # Fill the grid with frames
        for i in range(B):
            row = i // grid_w
            col = i % grid_w

            start_h = row * H
            end_h = start_h + H
            start_w = col * W
            end_w = start_w + W

            result[start_h:end_h, start_w:end_w] = frames[i]

        return result

    def __call__(
        self,
        env: EmbodiedEnv,
        env_ids: Union[torch.Tensor, None],
        name: str,
        resolution: tuple[int, int] = (640, 480),
        eye: tuple[float, float, float] = (0, 0, 2),
        target: tuple[float, float, float] = (0, 0, 0),
        up: tuple[float, float, float] = (0, 0, 1),
        intrinsics: tuple[float, float, float, float] = (
            600,
            600,
            320,
            240,
        ),
        max_env_num: int = 16,
        save_path: str = "./outputs/videos",
    ):
        # TODO: the current implementation will lost the final episode frames recording.
        # Check if the frames should be saved for the current episode
        if env.elapsed_steps.sum().item() == len(env_ids) and len(self._frames) > 0:
            video_name = f"episode_{self._current_episode}_{self._name}"
            images_to_video(self._frames, save_path, video_name, fps=20)

            self._current_episode += 1
            self._frames = []

        self.camera.update(fetch_only=self.camera.is_rt_enabled)
        data = self.camera.get_data()
        rgb = data["color"]

        num_frames = max(rgb.shape[0], max_env_num)
        rgb = rgb[:num_frames]
        rgb = self._draw_frames_into_one_image(rgb)[..., :3].cpu().numpy()
        self._frames.append(rgb)


class record_camera_data_async(record_camera_data):
    """Record camera data for multiple environments, merge and save as a single video at episode end."""

    def __init__(self, cfg: FunctorCfg, env: EmbodiedEnv):
        super().__init__(cfg, env)
        self._num_envs = min(4, getattr(env, "num_envs", 1))
        self._frames_list = [[] for _ in range(self._num_envs)]
        self._ep_idx = [0 for _ in range(self._num_envs)]

    def __call__(
        self,
        env: EmbodiedEnv,
        env_ids: Union[torch.Tensor, None],
        name: str,
        resolution: tuple[int, int] = (640, 480),
        eye: tuple[float, float, float] = (0, 0, 2),
        target: tuple[float, float, float] = (0, 0, 0),
        up: tuple[float, float, float] = (0, 0, 1),
        intrinsics: tuple[float, float, float, float] = (
            600,
            600,
            320,
            240,
        ),
        max_env_num: int = 16,
        save_path: str = "./outputs/videos",
    ):
        self.camera.update(fetch_only=self.camera.is_rt_enabled)
        data = self.camera.get_data()
        rgb = data["color"]  # shape: (num_envs, H, W, 4)
        if isinstance(rgb, torch.Tensor):
            rgb_np = rgb.cpu().numpy()
        else:
            rgb_np = rgb
        # Only collect frames for the first 4 environments
        for i in range(self._num_envs):
            self._frames_list[i].append(rgb_np[i][..., :])

        # Check if elapsed_steps==1 (just reset)
        elapsed = env.elapsed_steps
        if isinstance(elapsed, torch.Tensor):
            elapsed_np = elapsed.cpu().numpy()
        else:
            elapsed_np = elapsed
        # Only check reset for the first 4 environments
        ready_envs = [
            i
            for i in range(self._num_envs)
            if elapsed_np[i] == 1 and len(self._frames_list[i]) > 1
        ]
        # Used to temporarily store episode frames for each env
        if not hasattr(self, "_pending_env_episodes"):
            self._pending_env_episodes = {}
        for i in ready_envs:
            if i not in self._pending_env_episodes:
                self._pending_env_episodes[i] = self._frames_list[i][:-1]
                self._frames_list[i] = [
                    self._frames_list[i][-1]
                ]  # Only keep the first frame after reset
                self._ep_idx[i] += 1
        # If all specified envs have collected frames, concatenate and save
        if len(self._pending_env_episodes) == self._num_envs:
            min_len = min(len(frames) for frames in self._pending_env_episodes.values())
            big_frames = []
            for j in range(min_len):
                frames = [
                    self._pending_env_episodes[i][j] for i in range(self._num_envs)
                ]
                frames_tensor = torch.from_numpy(np.stack(frames)).to(torch.uint8)
                big_frame = (
                    self._draw_frames_into_one_image(frames_tensor)[..., :3]
                    .cpu()
                    .numpy()
                )
                big_frames.append(big_frame)
            video_name = f"ep{self._ep_idx[0]-1}_{self._name}_allenvs"
            images_to_video(big_frames, save_path, video_name, fps=20)
            self._pending_env_episodes.clear()


class validation_cameras(Functor):
    """
    This functor creates validation cameras during initialization and captures
    their data when called. The cameras are created once and reused for subsequent calls.
    """

    def __init__(self, cfg: FunctorCfg, env: EmbodiedEnv):
        super().__init__(cfg, env)
        # Store camera configurations
        self.cameras_cfg = cfg.params.get("cameras", [])
        # Create each camera in __init__
        self.camera_uids = []
        for cam_cfg in self.cameras_cfg:
            uid = cam_cfg.get("uid", "validation_camera")
            width = cam_cfg.get("width", 1280)
            height = cam_cfg.get("height", 960)
            enable_mask = cam_cfg.get("enable_mask", False)
            intrinsics = cam_cfg.get("intrinsics", [1400, 1400, 640, 480])
            extrinsics_cfg = cam_cfg.get("extrinsics", {})
            extrinsics = CameraCfg.ExtrinsicsCfg(**extrinsics_cfg)

            camera = env.sim.add_sensor(
                sensor_cfg=CameraCfg(
                    uid=uid,
                    width=width,
                    height=height,
                    enable_mask=enable_mask,
                    extrinsics=extrinsics,
                    intrinsics=intrinsics,
                )
            )
            if camera is not None:
                self.camera_uids.append(uid)

    def __call__(
        self,
        env: EmbodiedEnv,
        env_ids: Union[torch.Tensor, None],
    ):
        """Update cameras and return their data."""
        camera_data = {}
        for i, cam_uid in enumerate(self.camera_uids, start=1):
            camera = env.sim.get_sensor(cam_uid)
            camera.update()
            data = camera.get_data()
            camera_data[f"valid_rgb_{i}"] = data["color"]

        return camera_data



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

"""
This script demonstrates how to create a rigid object group using SimulationManager.
"""

import argparse
import time

from embodichain.lab.sim import SimulationManager, SimulationManagerCfg
from embodichain.lab.sim.cfg import RigidBodyAttributesCfg
from embodichain.lab.sim.shapes import CubeCfg
from embodichain.lab.sim.objects import (
    RigidObjectGroup,
    RigidObjectGroupCfg,
    RigidObjectCfg,
)


def main():
    """Main function to create and run the simulation scene."""

    # Parse command line arguments
    parser = argparse.ArgumentParser(
        description="Create a simulation scene with SimulationManager"
    )
    parser.add_argument(
        "--headless",
        action="store_true",
        default=False,
        help="Run simulation in headless mode",
    )
    parser.add_argument(
        "--num_envs", type=int, default=1, help="Number of parallel environments"
    )
    parser.add_argument(
        "--device", type=str, default="cpu", help="Simulation device (cuda or cpu)"
    )
    parser.add_argument(
        "--enable_rt",
        action="store_true",
        default=False,
        help="Enable ray tracing for better visuals",
    )
    args = parser.parse_args()

    # Configure the simulation
    sim_cfg = SimulationManagerCfg(
        width=1920,
        height=1080,
        headless=True,
        physics_dt=1.0 / 100.0,  # Physics timestep (100 Hz)
        sim_device=args.device,
        enable_rt=args.enable_rt,  # Enable ray tracing for better visuals
        num_envs=args.num_envs,
        arena_space=3.0,
    )

    # Create the simulation instance
    sim = SimulationManager(sim_cfg)

    physics_attrs = RigidBodyAttributesCfg(
        mass=1.0,
        dynamic_friction=0.5,
        static_friction=0.5,
        restitution=0.1,
    )

    # Add objects to the scene
    obj_group: RigidObjectGroup = sim.add_rigid_object_group(
        cfg=RigidObjectGroupCfg(
            uid="obj_group",
            rigid_objects={
                "cube_1": RigidObjectCfg(
                    uid="cube_1",
                    shape=CubeCfg(size=[0.1, 0.1, 0.1]),
                    attrs=physics_attrs,
                    init_pos=[0.0, 0.0, 1.0],
                ),
                "cube_2": RigidObjectCfg(
                    uid="cube_2",
                    shape=CubeCfg(size=[0.2, 0.2, 0.2]),
                    attrs=physics_attrs,
                    init_pos=[0.5, 0.0, 1.0],
                ),
                "cube_3": RigidObjectCfg(
                    uid="cube_3",
                    shape=CubeCfg(size=[0.3, 0.3, 0.3]),
                    attrs=physics_attrs,
                    init_pos=[-0.5, 0.0, 1.0],
                ),
            },
        )
    )

    print("[INFO]: Scene setup complete!")
    print(f"[INFO]: Running simulation with {args.num_envs} environment(s)")
    print("[INFO]: Press Ctrl+C to stop the simulation")

    # Open window when the scene has been set up
    if not args.headless:
        sim.open_window()

    # Run the simulation
    run_simulation(sim)


def run_simulation(sim: SimulationManager):
    """Run the simulation loop.

    Args:
        sim: The SimulationManager instance to run
    """

    # Initialize GPU physics if using CUDA
    if sim.is_use_gpu_physics:
        sim.init_gpu_physics()

    step_count = 0

    try:
        last_time = time.time()
        last_step = 0
        while True:
            # Update physics simulation
            sim.update(step=1)
            step_count += 1

            # Print FPS every second
            if step_count % 100 == 0:
                current_time = time.time()
                elapsed = current_time - last_time
                fps = (
                    sim.num_envs * (step_count - last_step) / elapsed
                    if elapsed > 0
                    else 0
                )
                print(f"[INFO]: Simulation step: {step_count}, FPS: {fps:.2f}")
                last_time = current_time
                last_step = step_count

    except KeyboardInterrupt:
        print("\n[INFO]: Stopping simulation...")
    finally:
        # Clean up resources
        sim.destroy()
        print("[INFO]: Simulation terminated successfully")


if __name__ == "__main__":
    main()



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------


import numpy as np
from typing import List, Union, TYPE_CHECKING
from embodichain.utils import logger

try:
    import pinocchio as pin
except ImportError:
    logger.log_warning("pinocchio not installed. Install with `pip install pin==2.7.0`")

try:
    from pink.configuration import Configuration
    from pink.tasks import Task
except ImportError:
    logger.log_warning(
        "pin-pink not installed. Install with `pip install pin-pink==3.4.0`"
    )


class NullSpacePostureTask(Task):
    r"""Pink-based task that adds a posture objective that is in the null space projection of other tasks.

    This task implements posture control in the null space of higher priority tasks
    (typically end-effector pose tasks) within the Pink inverse kinematics framework.

    **Mathematical Formulation:**

    For details on Pink Inverse Kinematics optimization formulation visit: https://github.com/stephane-caron/pink

    **Null Space Posture Task Implementation:**

    This task consists of two components:

    1. **Error Function**: The posture error is computed as:

    .. math::

        \mathbf{e}(\mathbf{q}) = \mathbf{M} \cdot (\mathbf{q}^* - \mathbf{q})

    where:
        - :math:`\mathbf{q}^*` is the target joint configuration
        - :math:`\mathbf{q}` is the current joint configuration
        - :math:`\mathbf{M}` is a joint selection mask matrix

    2. **Jacobian Matrix**: The task Jacobian is the null space projector:

    .. math::

        \mathbf{J}_{\text{posture}}(\mathbf{q}) = \mathbf{N}(\mathbf{q}) = \mathbf{I} - \mathbf{J}_{\text{primary}}^+ \mathbf{J}_{\text{primary}}

    where:
        - :math:`\mathbf{J}_{\text{primary}}` is the combined Jacobian of all higher priority tasks
        - :math:`\mathbf{J}_{\text{primary}}^+` is the pseudoinverse of the primary task Jacobian
        - :math:`\mathbf{N}(\mathbf{q})` is the null space projector matrix

    For example, if there are two frame tasks (e.g., controlling the pose of two end-effectors), the combined Jacobian
    :math:`\mathbf{J}_{\text{primary}}` is constructed by stacking the individual Jacobians for each frame vertically:

    .. math::

        \mathbf{J}_{\text{primary}} =
        \begin{bmatrix}
            \mathbf{J}_1(\mathbf{q}) \\
            \mathbf{J}_2(\mathbf{q})
        \end{bmatrix}

    where :math:`\mathbf{J}_1(\mathbf{q})` and :math:`\mathbf{J}_2(\mathbf{q})` are the Jacobians for the first and second frame tasks, respectively.

    The null space projector ensures that joint velocities in the null space produce zero velocity
    for the primary tasks: :math:`\mathbf{J}_{\text{primary}} \cdot \dot{\mathbf{q}}_{\text{null}} = \mathbf{0}`.

    **Task Integration:**

    When integrated into the Pink framework, this task contributes to the optimization as:

    .. math::

        \left\| \mathbf{N}(\mathbf{q}) \mathbf{v} + \mathbf{M} \cdot (\mathbf{q}^* - \mathbf{q}) \right\|_{W_{\text{posture}}}^2

    This formulation allows the robot to maintain a desired posture while respecting the constraints
    imposed by higher priority tasks (e.g., end-effector positioning).

    """

    def __init__(
        self,
        cost: float,
        lm_damping: float = 0.0,
        gain: float = 1.0,
        controlled_frames: list[str] | None = None,
        controlled_joints: list[str] | None = None,
    ) -> None:
        r"""Initialize the null space posture task.

        This task maintains a desired joint posture in the null space of higher-priority
        frame tasks. Joint selection allows excluding specific joints (e.g., wrist joints
        in humanoid manipulation) to prevent large rotational ranges from overwhelming
        errors in critical joints like shoulders and waist.

        Args:
            cost: Task weighting factor in the optimization objective.
                Units: :math:`[\text{cost}] / [\text{rad}]`.
            lm_damping: Levenberg-Marquardt regularization scale (unitless). Defaults to 0.0.
            gain: Task gain :math:`\alpha \in [0, 1]` for low-pass filtering.
                Defaults to 1.0 (no filtering).
            controlled_frames: Frame names whose Jacobians define the primary tasks for
                null space projection. If None or empty, no projection is applied.
            controlled_joints: Joint names to control in the posture task. If None or
                empty, all actuated joints are controlled.
        """
        super().__init__(cost=cost, gain=gain, lm_damping=lm_damping)
        self.target_q: np.ndarray | None = None
        self.controlled_frames: list[str] = controlled_frames or []
        self.controlled_joints: list[str] = controlled_joints or []
        self._joint_mask: np.ndarray | None = None
        self._frame_names: list[str] | None = None

    def __repr__(self) -> str:
        """Human-readable representation of the task."""
        return (
            f"NullSpacePostureTask(cost={self.cost}, gain={self.gain}, lm_damping={self.lm_damping},"
            f" controlled_frames={self.controlled_frames}, controlled_joints={self.controlled_joints})"
        )

    def _build_joint_mapping(self, configuration: Configuration) -> None:
        """Build joint mask and cache frequently used values.

        Creates a binary mask that selects which joints should be controlled
        in the posture task.

        Args:
            configuration: Robot configuration containing the model and joint information.
        """
        # Create joint mask for full configuration size
        self._joint_mask = np.zeros(configuration.model.nq)

        # Create dictionary for joint names to indices (exclude root joint)
        joint_names = configuration.model.names.tolist()[1:]

        # Build joint mask efficiently
        for i, joint_name in enumerate(joint_names):
            if joint_name in self.controlled_joints:
                self._joint_mask[i] = 1.0

        # Cache frame names for performance
        self._frame_names = list(self.controlled_frames)

    def set_target(self, target_q: np.ndarray) -> None:
        """Set target posture configuration.

        Args:
            target_q: Target vector in the configuration space. If the model
                has a floating base, then this vector should include
                floating-base coordinates (although they have no effect on the
                posture task since only actuated joints are controlled).
        """
        self.target_q = target_q.copy()

    def set_target_from_configuration(self, configuration: Configuration) -> None:
        """Set target posture from a robot configuration.

        Args:
            configuration: Robot configuration whose joint angles will be used
                as the target posture.
        """
        self.set_target(configuration.q)

    def compute_error(self, configuration: Configuration) -> np.ndarray:
        r"""Compute posture task error.

        The error computation follows:

        .. math::

            \mathbf{e}(\mathbf{q}) = \mathbf{M} \cdot (\mathbf{q}^* - \mathbf{q})

        where :math:`\mathbf{M}` is the joint selection mask and :math:`\mathbf{q}^* - \mathbf{q}`
        is computed using Pinocchio's difference function to handle joint angle wrapping.

        Args:
            configuration: Robot configuration :math:`\mathbf{q}`.

        Returns:
            Posture task error :math:`\mathbf{e}(\mathbf{q})` with the same dimension
            as the configuration vector, but with zeros for non-controlled joints.

        Raises:
            ValueError: If no posture target has been set.
        """
        if self.target_q is None:
            raise ValueError("No posture target has been set. Call set_target() first.")

        # Initialize joint mapping if needed
        if self._joint_mask is None:
            self._build_joint_mapping(configuration)

        # Compute configuration difference using Pinocchio's difference function
        # This handles joint angle wrapping correctly
        err = pin.difference(
            configuration.model,
            self.target_q,
            configuration.q,
        )

        # Apply pre-computed joint mask to select only controlled joints
        return self._joint_mask * err

    def compute_jacobian(self, configuration: Configuration) -> np.ndarray:
        r"""Compute the null space projector Jacobian.

        The null space projector is defined as:

        .. math::

            \mathbf{N}(\mathbf{q}) = \mathbf{I} - \mathbf{J}_{\text{primary}}^+ \mathbf{J}_{\text{primary}}

        where:
            - :math:`\mathbf{J}_{\text{primary}}` is the combined Jacobian of all controlled frames
            - :math:`\mathbf{J}_{\text{primary}}^+` is the pseudoinverse of the primary task Jacobian
            - :math:`\mathbf{I}` is the identity matrix

        The null space projector ensures that joint velocities in the null space produce
        zero velocity for the primary tasks: :math:`\mathbf{J}_{\text{primary}} \cdot \dot{\mathbf{q}}_{\text{null}} = \mathbf{0}`.

        If no controlled frames are specified, returns the identity matrix.

        Args:
            configuration: Robot configuration :math:`\mathbf{q}`.

        Returns:
            Null space projector matrix :math:`\mathbf{N}(\mathbf{q})` with dimensions
            :math:`n_q \times n_q` where :math:`n_q` is the number of configuration variables.
        """
        # Initialize joint mapping if needed
        if self._frame_names is None:
            self._build_joint_mapping(configuration)

        # If no frame tasks are defined, return identity matrix (no null space projection)
        if not self._frame_names:
            return np.eye(configuration.model.nq)

        # Get Jacobians for all frame tasks and combine them
        J_frame_tasks = [
            configuration.get_frame_jacobian(frame_name)
            for frame_name in self._frame_names
        ]
        J_combined = np.concatenate(J_frame_tasks, axis=0)

        # Compute null space projector: N = I - J^+ * J
        N_combined = (
            np.eye(J_combined.shape[1]) - np.linalg.pinv(J_combined) @ J_combined
        )

        return N_combined



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

"""Dataset manager for orchestrating dataset collection functors."""

from __future__ import annotations

import inspect
from typing import TYPE_CHECKING, Any, Dict, Optional, Union
from collections.abc import Sequence

import torch
from prettytable import PrettyTable

from embodichain.utils import logger
from .manager_base import ManagerBase
from .cfg import DatasetFunctorCfg

if TYPE_CHECKING:
    from embodichain.lab.gym.envs import EmbodiedEnv


class DatasetManager(ManagerBase):
    """Manager for orchestrating dataset collection and saving using functors.

    The dataset manager supports multiple dataset formats through a functor system:
    - LeRobot format (via LeRobotRecorder)
    - HDF5 format (via HDF5Recorder)
    - Zarr format (via ZarrRecorder)
    - Custom formats (via user-defined functors)

    Each functor's step() method is called once per environment step and handles:
    - Recording observation-action pairs
    - Detecting episode completion (dones=True)
    - Auto-saving completed episodes

    Example configuration:
        >>> from embodichain.lab.gym.envs.managers.cfg import DatasetFunctorCfg
        >>> from embodichain.lab.gym.envs.managers.datasets import LeRobotRecorder
        >>>
        >>> @configclass
        >>> class MyEnvCfg:
        >>>     dataset: dict = {
        >>>         "lerobot": DatasetFunctorCfg(
        >>>             func=LeRobotRecorder,
        >>>             params={
        >>>                 "robot_meta": {...},
        >>>                 "instruction": {"lang": "pick and place"},
        >>>                 "extra": {"scene_type": "kitchen"},
        >>>                 "save_path": "/data/datasets"
        >>>             }
        >>>         )
        >>>     }
    """

    _env: EmbodiedEnv
    """The environment instance."""

    def __init__(self, cfg: object, env: EmbodiedEnv):
        """Initialize the dataset manager.

        Args:
            cfg: Configuration object containing dataset functor configurations.
            env: The environment instance.
        """
        # Store functors by mode (similar to EventManager)
        self._mode_functor_names: dict[str, list[str]] = {}
        self._mode_functor_cfgs: dict[str, list[DatasetFunctorCfg]] = {}
        self._mode_class_functor_cfgs: dict[str, list[DatasetFunctorCfg]] = {}

        # Call base class to parse functors
        super().__init__(cfg, env)

        ## TODO: fix configurable_action.py to avoid getting env.metadata['dataset']
        # Extract robot_meta and instruction from functor params or plain config and add to env.metadata for backward compatibility
        # This allows legacy code (like action_bank) to access robot_meta via env.metadata["dataset"]["robot_meta"]
        robot_meta_found = False

        # First, try to extract from functor params
        for mode_cfgs in self._mode_functor_cfgs.values():
            for functor_cfg in mode_cfgs:
                if (
                    "robot_meta" in functor_cfg.params
                    or "instruction" in functor_cfg.params
                ):
                    if not hasattr(env, "metadata"):
                        env.metadata = {}
                    if "dataset" not in env.metadata:
                        env.metadata["dataset"] = {}
                    if "robot_meta" in functor_cfg.params:
                        env.metadata["dataset"]["robot_meta"] = functor_cfg.params[
                            "robot_meta"
                        ]
                    if "instruction" in functor_cfg.params:
                        env.metadata["dataset"]["instruction"] = functor_cfg.params[
                            "instruction"
                        ]
                    logger.log_info(
                        "Added robot_meta and instruction to env.metadata for backward compatibility"
                    )
                    robot_meta_found = True
                    break
            if robot_meta_found:
                break

        # If not found in functor params, try to extract from plain config
        if not robot_meta_found:
            # Check if config is dict or object
            if isinstance(self.cfg, dict):
                cfg_items = self.cfg.items()
            else:
                cfg_items = self.cfg.__dict__.items()

            for config_name, config_value in cfg_items:
                if config_name == "robot_meta" and isinstance(config_value, dict):
                    if not hasattr(env, "metadata"):
                        env.metadata = {}
                    if "dataset" not in env.metadata:
                        env.metadata["dataset"] = {}
                    env.metadata["dataset"]["robot_meta"] = config_value
                    logger.log_info(
                        "Added robot_meta to env.metadata for backward compatibility (from plain config)"
                    )
                    break

        logger.log_info(
            f"DatasetManager initialized with {sum(len(v) for v in self._mode_functor_names.values())} functors"
        )

    def __str__(self) -> str:
        """Returns: A string representation for dataset manager."""
        msg = f"<DatasetManager> contains {len(self._functor_names)} active functors.\n"

        table = PrettyTable()
        table.title = "Active Dataset Functors"
        table.field_names = ["Index", "Name", "Type"]
        table.align["Name"] = "l"

        for index, name in enumerate(self._functor_names):
            functor_cfg = self._functor_cfgs[index]
            functor_type = (
                functor_cfg.func.__class__.__name__
                if hasattr(functor_cfg.func, "__class__")
                else str(functor_cfg.func)
            )
            table.add_row([index, name, functor_type])

        msg += table.get_string()
        msg += "\n"

        return msg

    """
    Properties.
    """

    @property
    def active_functors(self) -> dict[str, list[str]]:
        """Name of active dataset functors by mode.

        The keys are the modes and the values are the names of the dataset functors.
        """
        return self._mode_functor_names

    @property
    def available_modes(self) -> list[str]:
        """List of available modes for the dataset manager."""
        return list(self._mode_functor_names.keys())

    """
    Operations.
    """

    def reset(
        self, env_ids: Union[Sequence[int], torch.Tensor, None] = None
    ) -> dict[str, float]:
        """Reset all dataset functors.

        Args:
            env_ids: The environment ids. Defaults to None.

        Returns:
            Empty dict (no logging info).
        """
        # Call reset on all functor instances across all modes
        for mode_cfgs in self._mode_functor_cfgs.values():
            for functor_cfg in mode_cfgs:
                if hasattr(functor_cfg.func, "reset"):
                    functor_cfg.func.reset(env_ids=env_ids)

        return {}

    def apply(
        self,
        mode: str,
        env_ids: Union[Sequence[int], torch.Tensor, None] = None,
    ) -> None:
        """Apply dataset functors for the specified mode.

        This method saves completed episodes by reading data from the environment's
        episode buffers. It should be called before clearing the buffers during reset.

        Args:
            mode: The mode to apply (currently only "save" is supported).
            env_ids: The indices of the environments to apply the functor to.
                Defaults to None, in which case the functor is applied to all environments.
        """
        # check if mode is valid
        if mode not in self._mode_functor_names:
            logger.log_warning(
                f"Dataset mode '{mode}' is not defined. Skipping dataset operation."
            )
            return

        # iterate over all the dataset functors for this mode
        for functor_cfg in self._mode_functor_cfgs[mode]:
            functor_cfg.func(
                self._env,
                env_ids,
                **functor_cfg.params,
            )

    def finalize(self) -> Optional[str]:
        """Finalize all dataset functors.

        Called when the environment is closed. Saves any remaining episodes
        and finalizes all datasets.

        Returns:
            Path to the first saved dataset, or None if failed.
        """
        dataset_paths = []

        # Call finalize on all functor instances across all modes
        for mode_cfgs in self._mode_functor_cfgs.values():
            for functor_cfg in mode_cfgs:
                if hasattr(functor_cfg.func, "finalize"):
                    try:
                        path = functor_cfg.func.finalize()
                        if path:
                            dataset_paths.append(path)
                    except Exception as e:
                        logger.log_error(f"Failed to finalize functor: {e}")

        if dataset_paths:
            logger.log_info(f"Finalized {len(dataset_paths)} datasets")
            return dataset_paths[0]

        return None

    def get_cached_data(self) -> list[Dict[str, Any]]:
        """Get cached data from all dataset functors (for online training).

        Iterates through all functors and collects cached data from those
        that support online training mode (have get_cached_data method).

        Returns:
            List of cached data dictionaries from all functors.
        """
        all_cached_data = []

        # Iterate through all modes and functors
        for mode_cfgs in self._mode_functor_cfgs.values():
            for functor_cfg in mode_cfgs:
                if hasattr(functor_cfg.func, "get_cached_data"):
                    cached_data = functor_cfg.func.get_cached_data()
                    all_cached_data.extend(cached_data)

        return all_cached_data

    def clear_cache(self) -> int:
        """Clear cached data from all dataset functors (for online training).

        Iterates through all functors and clears their cache if they
        support online training mode (have clear_cache method).

        Returns:
            Total number of cached items cleared across all functors.
        """
        total_cleared = 0

        # Iterate through all modes and functors
        for mode_cfgs in self._mode_functor_cfgs.values():
            for functor_cfg in mode_cfgs:
                if hasattr(functor_cfg.func, "clear_cache"):
                    cleared = functor_cfg.func.clear_cache()
                    total_cleared += cleared

        return total_cleared

    def get_functor_cfg(self, functor_name: str) -> DatasetFunctorCfg:
        """Gets the configuration for the specified functor.

        Args:
            functor_name: The name of the dataset functor.

        Returns:
            The configuration of the dataset functor.

        Raises:
            ValueError: If the functor name is not found.
        """
        for mode, functors in self._mode_functor_names.items():
            if functor_name in functors:
                return self._mode_functor_cfgs[mode][functors.index(functor_name)]
        logger.log_error(f"Dataset functor '{functor_name}' not found.")

    def _prepare_functors(self):
        """Prepare dataset functors from configuration.

        This method parses the configuration and initializes all dataset functors,
        organizing them by mode (similar to EventManager).
        """
        # Check if config is dict already
        if isinstance(self.cfg, dict):
            cfg_items = self.cfg.items()
        else:
            cfg_items = self.cfg.__dict__.items()

        # Iterate over all the functors
        for functor_name, functor_cfg in cfg_items:
            # Check for non config
            if functor_cfg is None:
                continue

            # Skip non-functor configurations (e.g., robot_meta which is a plain dict)
            # Functor configurations must have a "func" field
            if isinstance(functor_cfg, dict) and "func" not in functor_cfg:
                # This is a plain configuration (not a functor), skip it
                continue

            # Convert dict to DatasetFunctorCfg if needed (for JSON configs)
            if isinstance(functor_cfg, dict):
                functor_cfg = DatasetFunctorCfg(**functor_cfg)

            # Check for valid config type
            if not isinstance(functor_cfg, DatasetFunctorCfg):
                raise TypeError(
                    f"Configuration for '{functor_name}' is not of type DatasetFunctorCfg."
                    f" Received: '{type(functor_cfg)}'."
                )

            # Resolve common parameters
            # min_argc=7 to skip: env, env_ids, obs, action, dones, terminateds, info
            # These are runtime positional arguments, not config parameters
            self._resolve_common_functor_cfg(functor_name, functor_cfg, min_argc=7)

            # Check if mode is a new mode
            if functor_cfg.mode not in self._mode_functor_names:
                # add new mode
                self._mode_functor_names[functor_cfg.mode] = []
                self._mode_functor_cfgs[functor_cfg.mode] = []
                self._mode_class_functor_cfgs[functor_cfg.mode] = []

            # Add functor name and parameters
            self._mode_functor_names[functor_cfg.mode].append(functor_name)
            self._mode_functor_cfgs[functor_cfg.mode].append(functor_cfg)

            # Check if the functor is a class
            if inspect.isclass(functor_cfg.func):
                self._mode_class_functor_cfgs[functor_cfg.mode].append(functor_cfg)



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

from __future__ import annotations

import pytest
import torch

from embodichain.lab.gym.envs.managers import (
    ActionManager,
    DeltaQposTerm,
    EefPoseTerm,
    QposTerm,
    QposNormalizedTerm,
    QvelTerm,
    QfTerm,
)
from embodichain.lab.gym.envs.managers.cfg import ActionTermCfg


class MockEnv:
    """Minimal mock env for ActionTerm tests."""

    def __init__(self, num_envs: int = 4, action_dim: int = 6):
        self.num_envs = num_envs
        self.active_joint_ids = list(range(action_dim))
        self.device = torch.device("cpu")

    def get_qpos(self):
        return torch.zeros(
            self.num_envs, len(self.active_joint_ids), device=self.device
        )

    @property
    def robot(self):
        """DeltaQposTerm uses env.robot.get_qpos()."""
        return self


class MockEnvWithLimits(MockEnv):
    """Mock env with qpos_limits for QposNormalizedTerm."""

    def __init__(self, num_envs: int = 4, action_dim: int = 6):
        super().__init__(num_envs, action_dim)
        # qpos_limits shape: (1, dof, 2) for [low, high]
        self._qpos_limits = torch.zeros(1, action_dim, 2)
        self._qpos_limits[..., 0] = -1.0
        self._qpos_limits[..., 1] = 1.0

    @property
    def robot(self):
        return self

    @property
    def body_data(self):
        class BodyData:
            def __init__(_, limits):
                _.qpos_limits = limits

        return BodyData(self._qpos_limits)


class MockEnvForEef(MockEnv):
    """Mock env with compute_ik for EefPoseTerm."""

    def __init__(self, num_envs: int = 2, action_dim: int = 6):
        super().__init__(num_envs, action_dim)

    def compute_ik(self, pose, joint_seed):
        """Return (all success, joint_seed) to simulate IK success."""
        batch_size = joint_seed.shape[0]
        ret = torch.ones(batch_size, dtype=torch.bool, device=self.device)
        return ret, joint_seed.clone()


def test_delta_qpos_term_process_action():
    """DeltaQposTerm: qpos = current_qpos + scale * action."""
    env = MockEnv(num_envs=4, action_dim=6)
    cfg = ActionTermCfg(func=DeltaQposTerm, params={"scale": 0.1})
    term = DeltaQposTerm(cfg, env)

    action = torch.ones(4, 6) * 2.0
    result = term.process_action(action)

    assert "qpos" in result
    expected = env.get_qpos() + 0.1 * action
    torch.testing.assert_close(result["qpos"], expected)
    assert term.action_dim == 6


def test_qpos_term_process_action():
    """QposTerm: qpos = scale * action."""
    env = MockEnv(num_envs=2, action_dim=3)
    cfg = ActionTermCfg(func=QposTerm, params={"scale": 0.5})
    term = QposTerm(cfg, env)

    action = torch.ones(2, 3)
    result = term.process_action(action)

    assert "qpos" in result
    torch.testing.assert_close(result["qpos"], torch.ones(2, 3) * 0.5)
    assert term.action_dim == 3


def test_qpos_normalized_term_process_action():
    """QposNormalizedTerm: [-1,1] -> [low, high] with scale=1."""
    env = MockEnvWithLimits(num_envs=2, action_dim=2)
    cfg = ActionTermCfg(func=QposNormalizedTerm, params={"scale": 1.0})
    term = QposNormalizedTerm(cfg, env)

    # action=-1 -> low, action=1 -> high
    action = torch.tensor([[-1.0, -1.0], [1.0, 1.0]])
    result = term.process_action(action)

    assert "qpos" in result
    # low=-1, high=1: qpos = low + (action + 1.0) * 0.5 * (high - low)
    expected = torch.tensor([[-1.0, -1.0], [1.0, 1.0]])
    torch.testing.assert_close(result["qpos"], expected)
    assert term.action_dim == 2


def test_eef_pose_term_process_action_6d():
    """EefPoseTerm: 6D pose (x,y,z,euler) -> IK -> qpos."""
    env = MockEnvForEef(num_envs=2, action_dim=6)
    cfg = ActionTermCfg(func=EefPoseTerm, params={"scale": 1.0, "pose_dim": 6})
    term = EefPoseTerm(cfg, env)

    # 6D: position + euler angles
    action = torch.zeros(2, 6)
    action[:, :3] = 0.1  # position
    action[:, 3:6] = 0.0  # euler (identity rotation)
    result = term.process_action(action)

    assert "qpos" in result
    assert "ik_success" in result
    assert result["qpos"].shape == (2, 6)
    assert result["ik_success"].shape == (2,)
    # Mock returns joint_seed (zeros); verify output matches
    torch.testing.assert_close(result["qpos"], env.get_qpos())
    assert term.action_dim == 6


def test_eef_pose_term_process_action_7d():
    """EefPoseTerm: 7D pose (x,y,z,quat) -> IK -> qpos."""
    env = MockEnvForEef(num_envs=2, action_dim=6)
    cfg = ActionTermCfg(func=EefPoseTerm, params={"scale": 1.0, "pose_dim": 7})
    term = EefPoseTerm(cfg, env)

    # 7D: position + quaternion (w,x,y,z)
    action = torch.zeros(2, 7)
    action[:, :3] = 0.1
    action[:, 3] = 1.0  # quat w
    action[:, 4:7] = 0.0  # quat x,y,z (identity)
    result = term.process_action(action)

    assert "qpos" in result
    assert "ik_success" in result
    assert result["qpos"].shape == (2, 6)
    torch.testing.assert_close(result["qpos"], env.get_qpos())
    assert term.action_dim == 7


def test_eef_pose_term_invalid_dim_raises():
    """EefPoseTerm raises ValueError for non-6D/7D action."""
    env = MockEnvForEef(num_envs=2, action_dim=6)
    cfg = ActionTermCfg(func=EefPoseTerm, params={"scale": 1.0, "pose_dim": 5})
    term = EefPoseTerm(cfg, env)

    with pytest.raises(ValueError, match="EEF pose action must be 6D or 7D"):
        term.process_action(torch.zeros(2, 5))


def test_qvel_term_process_action():
    """QvelTerm: qvel = scale * action."""
    env = MockEnv(num_envs=2, action_dim=3)
    cfg = ActionTermCfg(func=QvelTerm, params={"scale": 0.2})
    term = QvelTerm(cfg, env)

    action = torch.ones(2, 3)
    result = term.process_action(action)

    assert "qvel" in result
    torch.testing.assert_close(result["qvel"], torch.ones(2, 3) * 0.2)


def test_qf_term_process_action():
    """QfTerm: qf = scale * action."""
    env = MockEnv(num_envs=2, action_dim=3)
    cfg = ActionTermCfg(func=QfTerm, params={"scale": 10.0})
    term = QfTerm(cfg, env)

    action = torch.ones(2, 3)
    result = term.process_action(action)

    assert "qf" in result
    torch.testing.assert_close(result["qf"], torch.ones(2, 3) * 10.0)


def test_action_manager_tensor_input():
    """ActionManager passes tensor to first (active) term."""
    env = MockEnv(num_envs=2, action_dim=3)
    cfg = {
        "delta_qpos": ActionTermCfg(func=DeltaQposTerm, params={"scale": 0.1}),
    }
    manager = ActionManager(cfg, env)

    action = torch.ones(2, 3)
    result = manager.process_action(action)

    assert "qpos" in result
    expected = env.get_qpos() + 0.1 * action
    torch.testing.assert_close(result["qpos"], expected)
    assert manager.action_type == "delta_qpos"
    assert manager.total_action_dim == 3


def test_action_manager_dict_input():
    """ActionManager uses key to select term for dict input."""
    env = MockEnv(num_envs=2, action_dim=3)
    cfg = {
        "delta_qpos": ActionTermCfg(func=DeltaQposTerm, params={"scale": 0.1}),
        "qpos": ActionTermCfg(func=QposTerm, params={"scale": 1.0}),
    }
    manager = ActionManager(cfg, env)

    action_dict = {"qpos": torch.ones(2, 3) * 0.5}
    result = manager.process_action(action_dict)

    assert "qpos" in result
    torch.testing.assert_close(result["qpos"], torch.ones(2, 3) * 0.5)


def test_action_manager_invalid_dict_raises():
    """ActionManager raises when dict has no matching term key."""
    env = MockEnv(num_envs=2, action_dim=3)
    cfg = {"delta_qpos": ActionTermCfg(func=DeltaQposTerm, params={"scale": 0.1})}
    manager = ActionManager(cfg, env)

    with torch.no_grad():
        with pytest.raises(ValueError, match="No valid action keys"):
            manager.process_action({"unknown_key": torch.ones(2, 3)})



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

import os
from dexsim.utility.path import get_resources_data_path
from embodichain.lab.sim import SimulationManager, SimulationManagerCfg
from embodichain.lab.sim.cfg import (
    SoftbodyVoxelAttributesCfg,
    SoftbodyPhysicalAttributesCfg,
)
from embodichain.lab.sim.shapes import MeshCfg
from embodichain.lab.sim.objects import (
    SoftObject,
    SoftObjectCfg,
)
import pytest

COW_PATH = get_resources_data_path("Model", "cow", "cow.obj")


class BaseSoftObjectTest:
    def setup_simulation(self):
        sim_cfg = SimulationManagerCfg(
            width=1920,
            height=1080,
            headless=True,
            physics_dt=1.0 / 100.0,  # Physics timestep (100 Hz)
            sim_device="cuda",
            enable_rt=False,  # Enable ray tracing for better visuals
            num_envs=4,
            arena_space=3.0,
        )

        # Create the simulation instance
        self.sim = SimulationManager(sim_cfg)

        assert os.path.isfile(COW_PATH)

        # Enable manual physics update for precise control
        self.n_envs = 4

        # add softbody to the scene
        self.cow: SoftObject = self.sim.add_soft_object(
            cfg=SoftObjectCfg(
                uid="cow",
                shape=MeshCfg(
                    fpath=get_resources_data_path("Model", "cow", "cow.obj"),
                ),
                init_pos=[0.0, 0.0, 3.0],
                voxel_attr=SoftbodyVoxelAttributesCfg(
                    simulation_mesh_resolution=8,
                    maximal_edge_length=0.5,
                ),
                physical_attr=SoftbodyPhysicalAttributesCfg(
                    youngs=1e6,
                    poissons=0.45,
                    density=100,
                    dynamic_friction=0.1,
                    min_position_iters=30,
                ),
            ),
        )

    def test_run_simulation(self):
        self.sim.init_gpu_physics()
        for _ in range(100):
            self.sim.update(step=1)
        self.cow.reset()
        for _ in range(100):
            self.sim.update(step=1)

    def test_remove(self):
        self.sim.remove_asset(self.cow.uid)
        assert (
            self.cow.uid not in self.sim._soft_objects
        ), "Cow UID still present after removal"

    def teardown_method(self):
        """Clean up resources after each test method."""
        self.sim.destroy()


class TestSoftObjectCUDA(BaseSoftObjectTest):
    def setup_method(self):
        self.setup_simulation()


if __name__ == "__main__":
    test = TestSoftObjectCUDA()
    test.setup_method()
    test.test_run_simulation()



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

import torch
import numpy as np

from embodichain.lab.gym.envs import EmbodiedEnv, EmbodiedEnvCfg
from embodichain.lab.gym.utils.registration import register_env
from embodichain.lab.sim.planners import (
    MotionGenerator,
    MotionGenCfg,
    MotionGenOptions,
    ToppraPlannerCfg,
    ToppraPlanOptions,
    PlanState,
    MoveType,
    MovePart,
)
from embodichain.lab.sim.planners.utils import TrajectorySampleMethod
from embodichain.utils import logger

__all__ = ["BlocksRankingRGBEnv"]


@register_env("BlocksRankingRGB-v1", max_episode_steps=600)
class BlocksRankingRGBEnv(EmbodiedEnv):
    def __init__(self, cfg: EmbodiedEnvCfg = None, **kwargs):
        super().__init__(cfg, **kwargs)

        action_config = kwargs.get("action_config", None)
        if action_config is not None:
            self.action_config = action_config

    def create_demo_action_list(self, *args, **kwargs):
        """
        Create a demonstration action list for ranking three blocks in RGB order from left to right.

        Now the expert trajectory follows the following strategy:
        - Do not move the green block block_2 (as the middle reference)
        - Right hand (right_arm + right_eef) takes the red block block_1 and places it on the left of the green block
        - Left hand (left_arm + left_eef) takes the blue block block_3 and places it on the right of the green block

        Returns:
            list: A list of demo actions (torch.Tensor) to be executed by env.step().
        """
        try:
            block1 = self.sim.get_rigid_object("block_1")  # Red
            block2 = self.sim.get_rigid_object("block_2")  # Green
            block3 = self.sim.get_rigid_object("block_3")  # Blue
        except Exception as e:
            logger.log_warning(f"Blocks not found: {e}, returning empty action list.")
            return []

        # Get block poses and positions
        b1_pose = block1.get_local_pose(to_matrix=True)
        b2_pose = block2.get_local_pose(to_matrix=True)
        b3_pose = block3.get_local_pose(to_matrix=True)
        b1_pos = b1_pose[:, :3, 3]
        b2_pos = b2_pose[:, :3, 3]
        b3_pos = b3_pose[:, :3, 3]

        # Construct the target line centered on the green block (Red < Green < Blue)
        base_x = b2_pos[:, 0]
        base_y = b2_pos[:, 1]
        base_z = b2_pos[:, 2]

        tgt_green = b2_pos  # not moved, kept for reference
        tgt_red = torch.stack([base_x - 0.12, base_y, base_z], dim=1)
        tgt_blue = torch.stack([base_x + 0.10, base_y, base_z], dim=1)

        # Right arm / right hand
        right_arm_ids = self.robot.get_joint_ids(name="right_arm")
        right_eef_ids = self.robot.get_joint_ids(name="right_eef")
        # Left arm / left hand
        left_arm_ids = self.robot.get_joint_ids(name="left_arm")
        left_eef_ids = self.robot.get_joint_ids(name="left_eef")

        init_qpos = self.robot.get_qpos()
        init_right_arm_qpos = init_qpos[:, right_arm_ids]
        init_right_arm_xpos = self.robot.compute_fk(
            qpos=init_right_arm_qpos, name="right_arm", to_matrix=True
        )
        init_left_arm_qpos = init_qpos[:, left_arm_ids]
        init_left_arm_xpos = self.robot.compute_fk(
            qpos=init_left_arm_qpos, name="left_arm", to_matrix=True
        )

        motion_cfg = MotionGenCfg(
            planner_cfg=ToppraPlannerCfg(
                robot_uid=self.robot.uid,
            )
        )
        self.motion_generator = MotionGenerator(cfg=motion_cfg)

        gripper_open = torch.tensor(
            [0.05, 0.05], dtype=torch.float32, device=self.device
        )
        gripper_close = torch.tensor(
            [0.0, 0.0], dtype=torch.float32, device=self.device
        )

        action_list = []

        def _ik_to_qpos(
            target_xpos: torch.Tensor, seed: torch.Tensor, arm_name: str, name: str
        ):
            is_success, qpos = self.robot.compute_ik(
                pose=target_xpos, joint_seed=seed, name=arm_name
            )
            success_flag = (
                is_success.all() if isinstance(is_success, torch.Tensor) else is_success
            )
            if not success_flag:
                logger.log_warning(f"IK failed for {name}, using previous qpos.")
                qpos = seed
            return qpos

        def _append_hold(
            qpos: torch.Tensor,
            num_steps: int,
            gripper_state: torch.Tensor,
            arm_ids,
            eef_ids,
        ):
            for _ in range(num_steps):
                action = init_qpos.clone()
                action[:, arm_ids] = qpos
                action[:, eef_ids] = gripper_state.unsqueeze(0).expand(
                    self.num_envs, -1
                )
                action_list.append(action)

        def _append_move_for_arm(
            qpos_start: torch.Tensor,
            qpos_end: torch.Tensor,
            num_steps: int,
            gripper_state: torch.Tensor,
            arm_ids,
            eef_ids,
            control_part: str,
        ):
            options = MotionGenOptions(
                control_part=control_part,
                plan_opts=ToppraPlanOptions(
                    constraints={
                        "velocity": 0.2,
                        "acceleration": 0.5,
                    },
                    sample_method=TrajectorySampleMethod.QUANTITY,
                    sample_interval=num_steps,
                ),
            )
            # Joint space trajectory
            target_states = [
                PlanState(qpos=qpos_start[0], move_type=MoveType.JOINT_MOVE),
                PlanState(
                    qpos=qpos_end[0],
                    move_type=MoveType.JOINT_MOVE,
                ),
            ]
            plan_result = self.motion_generator.generate(
                target_states=target_states, options=options
            )

            for qpos_item in plan_result.positions:
                qpos = torch.as_tensor(
                    qpos_item, dtype=torch.float32, device=self.device
                )
                qpos = qpos.flatten()
                if qpos.shape[0] != len(arm_ids):
                    logger.log_warning(
                        f"Qpos shape mismatch: got {qpos.shape[0]}, expected {len(arm_ids)}"
                    )
                    continue
                qpos = qpos.unsqueeze(0).expand(self.num_envs, -1)
                action = init_qpos.clone()
                action[:, arm_ids] = qpos
                action[:, eef_ids] = gripper_state.unsqueeze(0).expand(
                    self.num_envs, -1
                )
                action_list.append(action)

        def _pick_and_place(
            block_pos: torch.Tensor,
            place_pos: torch.Tensor,
            seed_qpos: torch.Tensor,
            init_arm_xpos: torch.Tensor,
            arm_ids,
            eef_ids,
            arm_name: str,
            tag: str,
        ):
            pick = init_arm_xpos.clone()
            pick[:, :3, 3] = block_pos + torch.tensor(
                [0.02, 0.0, -0.025], dtype=torch.float32, device=self.device
            ).unsqueeze(0)

            lift = pick.clone()
            lift[:, 2, 3] += 0.15

            place = init_arm_xpos.clone()
            place[:, :3, 3] = place_pos + torch.tensor(
                [0.025, 0.0, 0.02], dtype=torch.float32, device=self.device
            ).unsqueeze(0)

            # inverse kinematics for three key poses
            q_pick = _ik_to_qpos(pick, seed_qpos, arm_name, f"{tag}_pick")
            q_lift = _ik_to_qpos(lift, q_pick, arm_name, f"{tag}_lift")
            q_place = _ik_to_qpos(place, q_lift, arm_name, f"{tag}_place")

            # execution segments: seed -> pick(open gripper) -> lift -> place(close gripper)
            _append_move_for_arm(
                seed_qpos, q_pick, 20, gripper_open, arm_ids, eef_ids, arm_name
            )
            _append_hold(q_pick, 5, gripper_close, arm_ids, eef_ids)  # close gripper
            _append_move_for_arm(
                q_pick, q_lift, 20, gripper_close, arm_ids, eef_ids, arm_name
            )
            _append_move_for_arm(
                q_lift, q_place, 30, gripper_close, arm_ids, eef_ids, arm_name
            )
            _append_hold(q_place, 5, gripper_open, arm_ids, eef_ids)  # open gripper

            return q_place

        # 1. Right hand handles the red block
        current_seed_right = init_right_arm_qpos
        current_seed_right = _pick_and_place(
            b1_pos,
            tgt_red,
            current_seed_right,
            init_right_arm_xpos,
            right_arm_ids,
            right_eef_ids,
            "right_arm",
            "red",
        )

        init_qpos[:, right_arm_ids] = current_seed_right

        # 2. Left hand handles the blue block
        current_seed_left = init_left_arm_qpos
        current_seed_left = _pick_and_place(
            b3_pos,
            tgt_blue,
            current_seed_left,
            init_left_arm_xpos,
            left_arm_ids,
            left_eef_ids,
            "left_arm",
            "blue",
        )

        logger.log_info(f"Generated {len(action_list)} demo actions for RGB ranking")
        return action_list

    def is_task_success(self, **kwargs) -> torch.Tensor:
        """Determine if the task is successfully completed.

        The task is successful if:
        1. Three blocks are arranged in RGB order from front to back:
           - Red block (block_1) x < Green block (block_2) x < Blue block (block_3) x
        2. All blocks are close together (within tolerance)

        Args:
            **kwargs: Additional arguments for task-specific success criteria.

        Returns:
            torch.Tensor: A boolean tensor indicating success for each environment in the batch.
        """
        try:
            block1 = self.sim.get_rigid_object("block_1")  # Red
            block2 = self.sim.get_rigid_object("block_2")  # Green
            block3 = self.sim.get_rigid_object("block_3")  # Blue
        except Exception as e:
            logger.log_warning(f"Blocks not found: {e}, returning False.")
            return torch.zeros(self.num_envs, dtype=torch.bool, device=self.device)

        # Get block poses
        block1_pose = block1.get_local_pose(to_matrix=True)
        block2_pose = block2.get_local_pose(to_matrix=True)
        block3_pose = block3.get_local_pose(to_matrix=True)

        # Extract positions (x, y, z)
        block1_pos = block1_pose[:, :3, 3]  # (num_envs, 3)
        block2_pos = block2_pose[:, :3, 3]
        block3_pos = block3_pose[:, :3, 3]

        # Tolerance for checking if blocks are close together
        eps = torch.tensor([0.13, 0.03], dtype=torch.float32, device=self.device)

        # Check if blocks are close together in x-y plane
        # block1 and block2 should be close
        block1_block2_diff = torch.abs(block1_pos[:, :2] - block2_pos[:, :2])
        blocks_close_12 = torch.all(block1_block2_diff < eps.unsqueeze(0), dim=1)

        # block2 and block3 should be close
        block2_block3_diff = torch.abs(block2_pos[:, :2] - block3_pos[:, :2])
        blocks_close_23 = torch.all(block2_block3_diff < eps.unsqueeze(0), dim=1)

        # Check RGB order: block1 (red) x < block2 (green) x < block3 (blue) x
        rgb_order = (block1_pos[:, 0] < block2_pos[:, 0]) & (
            block2_pos[:, 0] < block3_pos[:, 0]
        )

        # Task succeeds if blocks are close together and in RGB order
        success = blocks_close_12 & blocks_close_23 & rgb_order

        return success



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

import numpy as np
import torch
from typing import Union

from embodichain.lab.sim.utility.workspace_analyzer.configs.sampling_config import (
    SamplingStrategy,
)

from embodichain.lab.sim.utility.workspace_analyzer.samplers.base_sampler import (
    BaseSampler,
)

__all__ = ["GaussianSampler"]


class GaussianSampler(BaseSampler):
    """Gaussian (Normal distribution) sampler.

    This sampler generates samples from a multivariate Gaussian (normal) distribution,
    which is useful for local workspace exploration around a specific point or for
    generating samples with controlled spread.

    Advantages:
        - Natural for local exploration
        - Concentrates samples near center
        - Smooth density falloff
        - Good for sensitivity analysis
        - Matches many physical phenomena

    Disadvantages:
        - May not cover entire space well
        - Can generate samples outside bounds (clipped)
        - Not uniform - biased toward center
        - Parameters require tuning

    Attributes:
        mean: Center point of the distribution (n_dims,). If None, uses bounds center.
        std: Standard deviation(s). Can be scalar (isotropic) or per-dimension (n_dims,).
        clip_to_bounds: Whether to clip samples to bounds. If False, rejects out-of-bounds.
    """

    def __init__(
        self,
        seed: int = 42,
        device: torch.device | None = None,
        mean: torch.Tensor | np.ndarray | None = None,
        std: Union[float, torch.Tensor, np.ndarray] = 0.3,
        clip_to_bounds: bool = True,
    ):
        """Initialize the Gaussian sampler.

        Args:
            seed: Random seed for reproducibility. Defaults to 42.
            device: PyTorch device (cpu/cuda). Defaults to cpu.
            mean: Mean of the distribution (center point). If None, uses midpoint of bounds.
                 Can be Tensor or ndarray of shape (n_dims,).
            std: Standard deviation(s). Defaults to 0.3.
                - Scalar: Isotropic (same std for all dimensions)
                - Tensor/array (n_dims,): Per-dimension std
            clip_to_bounds: Whether to clip samples to bounds. Defaults to True.
                           If False, regenerates out-of-bounds samples (slower but unbiased).

        """
        super().__init__(seed, device)
        self.mean = self._to_tensor(mean) if mean is not None else None

        if isinstance(std, (int, float)):
            self.std = float(std)
        else:
            self.std = self._to_tensor(std)

        self.clip_to_bounds = clip_to_bounds

    def _sample_from_bounds(
        self, bounds: torch.Tensor | np.ndarray, num_samples: int
    ) -> torch.Tensor:
        """Generate Gaussian-distributed samples within the given bounds.

        Args:
            bounds: Tensor/Array of shape (n_dims, 2) containing [lower, upper] bounds.
            num_samples: Number of samples to generate.

        Returns:
            Tensor of shape (num_samples, n_dims) containing the sampled points.

        Raises:
            ValueError: If bounds are invalid or num_samples is non-positive.

        Examples:
            >>> # Sample around center with std=0.2
            >>> sampler = GaussianSampler(seed=42, std=0.2, clip_to_bounds=True)
            >>> bounds = torch.tensor([[-1.0, 1.0], [-1.0, 1.0]], dtype=torch.float32)
            >>> samples = sampler.sample(bounds, num_samples=100)
            >>>
            >>> # Sample around custom point
            >>> mean = torch.tensor([0.5, 0.5], dtype=torch.float32)
            >>> sampler = GaussianSampler(mean=mean, std=0.15)
            >>> samples = sampler.sample(bounds, num_samples=100)
        """
        bounds = self._validate_bounds(bounds)

        if num_samples <= 0:
            raise ValueError(f"num_samples must be positive, got {num_samples}")

        n_dims = bounds.shape[0]

        # Determine mean (center of distribution)
        if self.mean is None:
            # Use center of bounds
            mean = (bounds[:, 0] + bounds[:, 1]) / 2.0
        else:
            mean = self.mean
            if mean.shape[0] != n_dims:
                raise ValueError(
                    f"Mean dimension ({mean.shape[0]}) doesn't match bounds ({n_dims})"
                )

        # Determine std
        if isinstance(self.std, float):
            std = torch.full((n_dims,), self.std, device=self.device)
        else:
            std = self.std
            if std.shape[0] != n_dims:
                raise ValueError(
                    f"Std dimension ({std.shape[0]}) doesn't match bounds ({n_dims})"
                )

        if self.clip_to_bounds:
            # Generate samples and clip
            samples = self._generate_and_clip(mean, std, bounds, num_samples)
        else:
            # Rejection sampling to avoid clipping bias
            samples = self._generate_with_rejection(mean, std, bounds, num_samples)

        # Validate samples
        self._validate_samples(samples, bounds)

        return samples

    def _generate_and_clip(
        self,
        mean: torch.Tensor,
        std: torch.Tensor,
        bounds: torch.Tensor,
        num_samples: int,
    ) -> torch.Tensor:
        """Generate Gaussian samples and clip to bounds.

        Args:
            mean: Mean vector (n_dims,).
            std: Standard deviation vector (n_dims,).
            bounds: Bounds tensor (n_dims, 2).
            num_samples: Number of samples to generate.

        Returns:
            Clipped samples (num_samples, n_dims).
        """
        # Generate Gaussian samples
        samples = torch.randn(num_samples, mean.shape[0], device=self.device)
        samples = mean + samples * std

        # Clip to bounds
        lower_bounds = bounds[:, 0]
        upper_bounds = bounds[:, 1]
        samples = torch.clamp(samples, lower_bounds, upper_bounds)

        return samples

    def _generate_with_rejection(
        self,
        mean: torch.Tensor,
        std: torch.Tensor,
        bounds: torch.Tensor,
        num_samples: int,
    ) -> torch.Tensor:
        """Generate Gaussian samples with rejection sampling for out-of-bounds.

        This avoids the bias introduced by clipping but is slower.

        Args:
            mean: Mean vector (n_dims,).
            std: Standard deviation vector (n_dims,).
            bounds: Bounds tensor (n_dims, 2).
            num_samples: Number of samples to generate.

        Returns:
            Valid samples within bounds (num_samples, n_dims).
        """
        lower_bounds = bounds[:, 0]
        upper_bounds = bounds[:, 1]

        accepted_samples = []
        max_iterations = 100
        iteration = 0

        while len(accepted_samples) < num_samples and iteration < max_iterations:
            # Generate more samples than needed to reduce iterations
            num_needed = num_samples - len(accepted_samples)
            num_generate = max(num_needed * 2, 100)  # Generate 2x to reduce rejections

            # Generate Gaussian samples
            samples = torch.randn(num_generate, mean.shape[0], device=self.device)
            samples = mean + samples * std

            # Check which samples are within bounds
            within_bounds = torch.all(
                (samples >= lower_bounds) & (samples <= upper_bounds), dim=1
            )

            valid_samples = samples[within_bounds]
            accepted_samples.append(valid_samples)

            iteration += 1

        if (
            len(accepted_samples) == 0
            or sum(s.shape[0] for s in accepted_samples) < num_samples
        ):
            raise RuntimeError(
                f"Failed to generate {num_samples} samples within bounds after {max_iterations} iterations. "
                "Consider using clip_to_bounds=True or adjusting mean/std parameters."
            )

        # Concatenate and trim to exact size
        all_samples = torch.cat(accepted_samples, dim=0)
        return all_samples[:num_samples]

    def get_strategy_name(self) -> str:
        """Get the name of the sampling strategy.

        Returns:
            String identifier for the sampling strategy.
        """
        return SamplingStrategy.GAUSSIAN.value

    def __repr__(self) -> str:
        """String representation of the sampler."""
        mean_str = "auto" if self.mean is None else f"shape={self.mean.shape}"
        std_str = (
            f"{self.std}" if isinstance(self.std, float) else f"shape={self.std.shape}"
        )
        return (
            f"{self.__class__.__name__}("
            f"strategy={self.get_strategy_name()}, "
            f"mean={mean_str}, "
            f"std={std_str}, "
            f"clip={self.clip_to_bounds}, "
            f"seed={self.seed})"
        )



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# All rights reserved.
# ----------------------------------------------------------------------------

from typing import Dict, Optional
from embodichain.lab.gym.envs import EmbodiedEnv, EmbodiedEnvCfg
from embodichain.lab.gym.utils.registration import register_env
from embodichain.lab.gym.envs.tasks.tableware.base_agent_env import BaseAgentEnv

__all__ = ["RearrangementEnv", "RearrangementAgentEnv"]


@register_env("Rearrangement-v3", max_episode_steps=600)
class RearrangementEnv(EmbodiedEnv):
    def __init__(self, cfg: EmbodiedEnvCfg = None, **kwargs):
        super().__init__(cfg, **kwargs)

        action_config = kwargs.get("action_config", None)
        if action_config is not None:
            self.action_config = action_config

    def is_task_success(self) -> bool:
        fork = self.sim.get_rigid_object("fork")
        spoon = self.sim.get_rigid_object("spoon")
        plate = self.sim.get_rigid_object("plate")
        plate_pose = plate.get_local_pose(to_matrix=True)
        # TODO: now only for 1 env
        (
            spoon_place_target_x,
            spoon_place_target_y,
            spoon_place_target_z,
        ) = self.affordance_datas["spoon_place_pose"][:3, 3]
        (
            fork_place_target_x,
            fork_place_target_y,
            fork_place_target_z,
        ) = self.affordance_datas["fork_place_pose"][:3, 3]

        spoon_pose = spoon.get_local_pose(to_matrix=True)
        spoon_x, spoon_y, spoon_z = spoon_pose[0, :3, 3]

        fork_pose = fork.get_local_pose(to_matrix=True)
        fork_x, fork_y, fork_z = fork_pose[0, :3, 3]

        tolerance = self.metadata.get("success_params", {}).get("tolerance", 0.02)

        # spoon and fork should with the x y range of tolerance related to plate.
        return ~(
            abs(spoon_x - spoon_place_target_x) > tolerance
            or abs(spoon_y - spoon_place_target_y) > tolerance
            or abs(fork_x - fork_place_target_x) > tolerance
            or abs(fork_y - fork_place_target_y) > tolerance
        )


@register_env("RearrangementAgent-v3", max_episode_steps=600)
class RearrangementAgentEnv(BaseAgentEnv, RearrangementEnv):
    def __init__(self, cfg: EmbodiedEnvCfg = None, **kwargs):
        super().__init__(cfg, **kwargs)
        super()._init_agents(**kwargs)

    def reset(self, seed: Optional[int] = None, options: Optional[Dict] = None):
        obs, info = super().reset(seed=seed, options=options)
        super().get_states()
        return obs, info

    def is_task_success(self):
        fork = self.sim.get_rigid_object("fork")
        spoon = self.sim.get_rigid_object("spoon")
        plate = self.sim.get_rigid_object("plate")

        plate_pose = plate.get_local_pose(to_matrix=True)
        spoon_place_target_y = plate_pose[0, 1, 3] - 0.16
        fork_place_target_y = plate_pose[0, 1, 3] + 0.16

        spoon_pose = spoon.get_local_pose(to_matrix=True)
        spoon_y = spoon_pose[0, 1, 3]

        fork_pose = fork.get_local_pose(to_matrix=True)
        fork_y = fork_pose[0, 1, 3]

        tolerance = self.metadata.get("success_params", {}).get("tolerance", 0.02)

        # spoon and fork should with the y range of tolerance related to plate.
        return (
            abs(spoon_y - spoon_place_target_y) <= tolerance
            and abs(fork_y - fork_place_target_y) <= tolerance
        )



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

import os
import torch
import pytest
import numpy as np

from embodichain.lab.sim import SimulationManager, SimulationManagerCfg
from embodichain.lab.sim.objects import Robot
from embodichain.lab.sim.cfg import RobotCfg
from embodichain.data import get_data_path


# Base test class for CPU and CUDA
class BaseSolverTest:
    sim = None  # Define as a class attribute

    def setup_simulation(self, solver_type: str):
        # Set up simulation with specified device (CPU or CUDA)
        config = SimulationManagerCfg(headless=True, sim_device="cpu")
        self.sim = SimulationManager(config)

        # Load robot URDF file
        urdf = get_data_path("DexforceW1V021/DexforceW1_v02_1.urdf")
        assert os.path.isfile(urdf)

        cfg_dict = {
            "fpath": urdf,
            "control_parts": {
                "left_arm": [f"LEFT_J{i+1}" for i in range(7)],
                "right_arm": [f"RIGHT_J{i+1}" for i in range(7)],
            },
            "solver_cfg": {
                "left_arm": {
                    "class_type": solver_type,
                    "end_link_name": "left_ee",
                    "root_link_name": "left_arm_base",
                    "ik_nearest_weight": [1.0, 1.0, 1.0, 0.9, 0.9, 0.1, 0.1],
                },
                "right_arm": {
                    "class_type": solver_type,
                    "end_link_name": "right_ee",
                    "root_link_name": "right_arm_base",
                },
            },
        }

        self.robot: Robot = self.sim.add_robot(cfg=RobotCfg.from_dict(cfg_dict))

        # Wait for robot to stabilize.
        self.sim.update(step=100)

    @pytest.mark.parametrize("arm_name", ["left_arm", "right_arm"])
    def test_ik(self, arm_name: str):
        # Test inverse kinematics (IK) with a 1x4x4 homogeneous matrix pose and a joint_seed

        qpos_fk = torch.tensor(
            [[0.0, 0.0, 0.0, -np.pi / 4, 0.0, 0.0, 0.0]], dtype=torch.float32
        )

        fk_xpos = self.robot.compute_fk(qpos=qpos_fk, name=arm_name, to_matrix=True)

        res, ik_qpos = self.robot.compute_ik(pose=fk_xpos, name=arm_name)

        if ik_qpos.dim() == 3:
            ik_xpos = self.robot.compute_fk(
                qpos=ik_qpos[0][0], name=arm_name, to_matrix=True
            )
        else:
            ik_xpos = self.robot.compute_fk(qpos=ik_qpos, name=arm_name, to_matrix=True)

        assert torch.allclose(
            fk_xpos, ik_xpos, atol=1e-2, rtol=1e-2
        ), f"FK and IK results do not match for {arm_name}"

        # test for failed xpos
        invalid_pose = torch.tensor(
            [
                [
                    [1.0, 0.0, 0.0, 10.0],
                    [0.0, 1.0, 0.0, 10.0],
                    [0.0, 0.0, 1.0, 10.0],
                    [0.0, 0.0, 0.0, 1.0],
                ]
            ],
            dtype=torch.float32,
            device=self.robot.device,
        )
        res, ik_qpos = self.robot.compute_ik(
            pose=invalid_pose, joint_seed=ik_qpos, name=arm_name
        )
        dof = ik_qpos.shape[-1]
        assert res[0] == False
        assert ik_qpos.shape == (1, dof)

    def teardown_method(self):
        """Clean up resources after each test method."""
        self.sim.destroy()


class TestPytorchSolver(BaseSolverTest):
    def setup_method(self):
        self.setup_simulation(solver_type="PytorchSolver")


if __name__ == "__main__":
    np.set_printoptions(precision=5, suppress=True)
    test_solver = TestPytorchSolver()
    test_solver.setup_method()



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

from dataclasses import dataclass
from typing import List
from enum import Enum


class MetricType(Enum):
    """Types of workspace metrics."""

    REACHABILITY = "reachability"
    """Measures reachable workspace volume and coverage."""

    MANIPULABILITY = "manipulability"
    """Measures dexterity/manipulability throughout workspace."""

    DENSITY = "density"
    """Measures point density and workspace distribution."""

    ALL = "all"
    """Compute all available metrics."""


@dataclass
class ReachabilityConfig:
    """Configuration for reachability metric."""

    voxel_size: float = 0.01
    """Size of voxels for volume calculation in meters."""

    min_points_per_voxel: int = 1
    """Minimum number of points in a voxel to consider it reachable."""

    compute_coverage: bool = True
    """Whether to compute coverage percentage relative to bounding box."""


@dataclass
class ManipulabilityConfig:
    """Configuration for manipulability metric."""

    jacobian_threshold: float = 0.01
    """Minimum manipulability value to consider valid."""

    compute_isotropy: bool = True
    """Whether to compute isotropy index (condition number)."""

    compute_heatmap: bool = False
    """Whether to generate manipulability heatmap."""


@dataclass
class DensityConfig:
    """Configuration for density metric."""

    radius: float = 0.05
    """Radius for local density estimation in meters."""

    k_neighbors: int = 30
    """Number of nearest neighbors for density estimation."""

    compute_distribution: bool = True
    """Whether to compute density distribution statistics."""


@dataclass
class MetricConfig:
    """Configuration for workspace analysis metrics."""

    enabled_metrics: List[MetricType] = None
    """List of metrics to compute. If None, computes all metrics."""

    reachability: ReachabilityConfig = None
    """Configuration for reachability metric."""

    manipulability: ManipulabilityConfig = None
    """Configuration for manipulability metric."""

    density: DensityConfig = None
    """Configuration for density metric."""

    save_results: bool = True
    """Whether to save metric results to file."""

    output_format: str = "json"
    """Output format for metrics. Options: 'json', 'yaml', 'pkl'."""

    def __post_init__(self):
        """Initialize sub-configs with defaults if not provided."""
        if self.enabled_metrics is None:
            self.enabled_metrics = [MetricType.ALL]

        if self.reachability is None:
            self.reachability = ReachabilityConfig()

        if self.manipulability is None:
            self.manipulability = ManipulabilityConfig()

        if self.density is None:
            self.density = DensityConfig()



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

import torch

from typing import Union, Tuple, List, TYPE_CHECKING
from dataclasses import MISSING
from copy import deepcopy

from embodichain.utils import configclass, logger
from embodichain.lab.sim.solvers import SolverCfg, BaseSolver
from embodichain.lab.sim.solvers.qpos_seed_sampler import QposSeedSampler
from embodichain.lab.sim.utility.solver_utils import validate_iteration_params

if TYPE_CHECKING:
    from typing import Self

from embodichain.lab.sim.utility.import_utils import (
    lazy_import_pytorch_kinematics,
)


@configclass
class PytorchSolverCfg(SolverCfg):
    """Configuration for the pytorch kinematics solver used in the robot simulation.

    This configuration includes properties related to the solver setup, such as the URDF path,
    the end link name, and the root link name, along with the Tool Center Point (TCP).
    """

    class_type: str = "PytorchSolver"

    # Solver iteration parameters
    pos_eps: float = 5e-4
    """Tolerance for convergence for position"""

    rot_eps: float = 5e-4
    """Tolerance for convergence for rotation"""

    max_iterations: int = 500
    """Maximum number of iterations for the solver"""

    dt: float = 0.1
    """Time step for numerical integration"""

    damp: float = 1e-6
    """Damping factor to prevent numerical instability"""

    is_only_position_constraint: bool = False
    """Flag to indicate whether the solver should only consider position constraints."""

    num_samples: int = 5
    """Number of samples to generate different joint seeds for IK iterations.

    A higher number of samples increases the chances of finding a valid solution
    """

    ik_nearest_weight: list[float] | None = None
    """Weights for the inverse kinematics nearest calculation.
    
    The weights influence how the solver prioritizes closeness to the seed position
    when multiple solutions are available.
    """

    def init_solver(
        self, device: torch.device = torch.device("cpu"), **kwargs
    ) -> "PytorchSolver":
        """Initialize the solver with the configuration.

        Args:
            device (torch.device): The device to use for the solver. Defaults to CPU.
            **kwargs: Additional keyword arguments that may be used for solver initialization.

        Returns:
            PytorchSolver: An initialized solver instance.
        """

        solver = PytorchSolver(cfg=self, device=device, **kwargs)

        # Set the Tool Center Point (TCP) for the solver
        solver.set_tcp(self._get_tcp_as_numpy())

        return solver


def ensure_pose_shape(func):
    """
    Decorator to ensure the input target_pose is of shape (n, 4, 4).
    If input is (4, 4), it will be converted to (1, 4, 4).
    Raises ValueError if shape is invalid.
    """

    def wrapper(self, target_xpos, *args, **kwargs):
        target_xpos = torch.as_tensor(
            target_xpos, device=self.device, dtype=torch.float32
        )
        if target_xpos.dim() == 2:
            if target_xpos.shape != (4, 4):
                raise ValueError("target_xpos must be of shape (4, 4) or (n, 4, 4).")
            target_xpos = target_xpos.unsqueeze(0)
        elif target_xpos.dim() == 3:
            if target_xpos.shape[1:] != (4, 4):
                raise ValueError("target_xpos must be of shape (4, 4) or (n, 4, 4).")
        else:
            raise ValueError(
                "target_xpos must be a tensor of shape (4, 4) or (n, 4, 4)."
            )
        return func(self, target_xpos, *args, **kwargs)

    return wrapper


class PytorchSolver(BaseSolver):
    def __init__(
        self,
        cfg: PytorchSolverCfg,
        device: str = None,
        **kwargs,
    ):
        r"""Initializes the PyTorch kinematics solver.

            This constructor sets up the kinematics solver using PyTorch,
            allowing for efficient computation of robot kinematics based on
            the specified URDF model.

        Args:
            cfg: The configuration for the solver.
            device (str, optional): The device to use for the solver (e.g., "cpu" or "cuda").
            **kwargs: Additional keyword arguments passed to the base solver.

        """
        super().__init__(cfg=cfg, device=device, **kwargs)

        self.pk = lazy_import_pytorch_kinematics()

        # Initialize solver parameters from configuration
        self._pos_eps = cfg.pos_eps
        self._rot_eps = cfg.rot_eps
        self._max_iterations = cfg.max_iterations
        self._dt = cfg.dt
        self._damp = cfg.damp
        self._is_only_position_constraint = cfg.is_only_position_constraint
        self._num_samples = cfg.num_samples

        # Get agent joint limits.
        self.lim = torch.tensor(
            self.pk_serial_chain.get_joint_limits(), device=self.device
        )

        # Inverse kinematics is available via damped least squares (iterative steps with Jacobian pseudo-inverse damped to avoid oscillation near singularlities).
        self.pik = self.pk.PseudoInverseIK(
            self.pk_serial_chain,
            pos_tolerance=self._pos_eps,
            rot_tolerance=self._rot_eps,
            joint_limits=self.lim.T,
            early_stopping_any_converged=True,
            max_iterations=self._max_iterations,
            lr=self._dt,
            num_retries=1,
        )

        self.dof = self.pk_serial_chain.n_joints

        self.upper_position_limits = self.pk_serial_chain.high
        self.lower_position_limits = self.pk_serial_chain.low

    def get_iteration_params(self) -> dict:
        r"""Returns the current iteration parameters.

        Returns:
            dict: A dictionary containing the current values of:
                - pos_eps (float): Pos convergence threshold
                - rot_eps (float): Rot convergence threshold
                - max_iterations (int): Maximum number of iterations.
                - dt (float): Time step size.
                - damp (float): Damping factor.
                - num_samples (int): Number of samples.
                - is_only_position_constraint (bool): Flag to indicate whether the solver should only consider position constraints.
        """
        return {
            "pos_eps": self._pos_eps,
            "rot_eps": self._rot_eps,
            "max_iterations": self._max_iterations,
            "dt": self._dt,
            "damp": self._damp,
            "num_samples": self._num_samples,
        }

    def set_iteration_params(
        self,
        pos_eps: float = 5e-4,
        rot_eps: float = 5e-4,
        max_iterations: int = 1000,
        dt: float = 0.1,
        damp: float = 1e-6,
        num_samples: int = 30,
        is_only_position_constraint: bool = False,
    ) -> bool:
        r"""Sets the iteration parameters for the kinematics solver.

        Args:
            pos_eps (float): Pos convergence threshold, must be positive.
            rot_eps (float): Rot convergence threshold, must be positive.
            max_iterations (int): Maximum number of iterations, must be positive.
            dt (float): Time step size, must be positive.
            damp (float): Damping factor, must be non-negative.
            num_samples (int): Number of samples, must be positive.
            is_only_position_constraint (bool): Flag to indicate whether the solver should only consider position constraints.

        Returns:
            bool: True if all parameters are valid and set, False otherwise.
        """
        # Validate parameters
        if not validate_iteration_params(
            pos_eps, rot_eps, max_iterations, dt, damp, num_samples
        ):
            return False

        # Set parameters if all are valid
        self._pos_eps = pos_eps
        self._rot_eps = rot_eps
        self._max_iterations = max_iterations
        self._dt = dt
        self._damp = damp
        self._num_samples = num_samples
        self._is_only_position_constraint = is_only_position_constraint

        self.pik = self.pk.PseudoInverseIK(
            self.pk_serial_chain,
            pos_tolerance=self._pos_eps,
            rot_tolerance=self._rot_eps,
            joint_limits=self.lim.T,
            early_stopping_any_converged=True,
            max_iterations=self._max_iterations,
            lr=self._dt,
            num_retries=1,
        )

        return True

    def _compute_inverse_kinematics(
        self, target_pose: torch.Tensor, joint_seed: torch.Tensor
    ) -> Tuple[Union[bool, torch.Tensor], torch.Tensor]:
        r"""Computes the inverse kinematics solutions for the given target poses and joint seeds.

        Args:
            target_pose (torch.Tensor): The target poses represented as a (batch_size, 4, 4) tensor.
            joint_seed (torch.Tensor): The initial joint positions used as a seed. It can be either a 1D tensor of shape (dof,) or a 2D tensor of shape (batch_size, dof).

        Returns:
            Tuple[Union[bool, torch.Tensor], torch.Tensor]:
                - First element:
                    - If solutions exist: torch.BoolTensor of shape (batch_size,) indicating convergence per pose
                    - If no solutions: Python False
                - Second element:
                    - If solutions exist: torch.Tensor of shape (batch_size, dof) containing joint solutions
                    - If no solutions: Empty torch.Tensor
        """
        target_pose = target_pose.to(self.device).float()
        joint_seed = joint_seed.to(self.device).float()

        # Extract translation and rotation parts
        pos = target_pose[:, :3, 3]
        rot = target_pose[:, :3, :3]

        tf = self.pk.Transform3d(
            pos=pos,
            rot=rot,
            device=self.device,
        )
        self.pik.initial_config = joint_seed

        result = self.pik.solve(tf)

        if result.converged_any.any().item():
            return result.converged_any, result.solutions[:, 0, :].squeeze(0)

        return False, torch.empty(0)

    @staticmethod
    def _qpos_to_limits_single(
        q: torch.Tensor,
        joint_seed: torch.Tensor,
        lower_position_limits: torch.Tensor,
        upper_position_limits: torch.Tensor,
        ik_nearest_weight: torch.Tensor,
        periodic_mask: torch.Tensor = None,  # Optional mask for periodic joints
    ) -> torch.Tensor:
        """
        Adjusts the given joint positions (q) to fit within the specified limits while minimizing the difference to the seed position.

        Args:
            q (torch.Tensor): The initial joint positions.
            joint_seed (torch.Tensor): The seed joint positions for comparison.
            lower_position_limits (torch.Tensor): The lower bounds for the joint positions.
            upper_position_limits (torch.Tensor): The upper bounds for the joint positions.
            ik_nearest_weight (torch.Tensor): The weights for the inverse kinematics nearest calculation.
            periodic_mask (torch.Tensor, optional): Boolean mask indicating which joints are periodic.

        Returns:
            torch.Tensor: The adjusted joint positions that fit within the limits.
        """
        device = q.device
        joint_seed = joint_seed.to(device)
        lower = lower_position_limits.to(device)
        upper = upper_position_limits.to(device)
        weight = ik_nearest_weight.to(device)

        # If periodic_mask is not provided, assume all joints are periodic
        if periodic_mask is None:
            periodic_mask = torch.ones_like(q, dtype=torch.bool, device=device)

        # Only enumerate [-2π, 0, 2π] for periodic joints, single value for non-periodic
        offsets = torch.tensor([-2 * torch.pi, 0, 2 * torch.pi], device=device)
        candidate_list = []
        for i in range(q.size(0)):
            if periodic_mask[i]:
                candidate_list.append(q[i] + offsets)
            else:
                candidate_list.append(q[i].unsqueeze(0))
        # Generate all possible combinations
        mesh = torch.meshgrid(*candidate_list, indexing="ij")
        candidates = torch.stack([m.reshape(-1) for m in mesh], dim=1)
        # Filter candidates that are out of limits
        mask = (candidates >= lower) & (candidates <= upper)
        valid_mask = mask.all(dim=1)
        valid_candidates = candidates[valid_mask]
        if valid_candidates.shape[0] == 0:
            return torch.tensor([]).to(device)
        # Compute weighted distance to seed and select the closest
        diffs = torch.abs(valid_candidates - joint_seed) * weight
        distances = torch.sum(diffs, dim=1)
        min_idx = torch.argmin(distances)
        return valid_candidates[min_idx]

    def _qpos_to_limits(
        self, qpos_list_split: torch.Tensor, joint_seed: torch.Tensor
    ) -> torch.Tensor:
        r"""Adjusts a batch of joint positions to fit within joint limits and minimize the weighted distance to the seed position.

        Args:
            qpos_list_split (torch.Tensor): Batch of candidate joint positions, shape (N, dof).
            joint_seed (torch.Tensor): The reference joint positions for comparison, shape (dof,).

        Returns:
            torch.Tensor: Batch of adjusted joint positions that fit within the limits, shape (M, dof),
                        where M <= N (invalid candidates are filtered out).
        """

        periodic_mask = torch.ones_like(
            qpos_list_split[0], dtype=torch.bool, device=self.device
        )

        adjusted_qpos_list = [
            self._qpos_to_limits_single(
                q,
                joint_seed,
                self.lower_position_limits,
                self.upper_position_limits,
                self.ik_nearest_weight,
                periodic_mask,
            )
            for q in qpos_list_split
        ]

        # Filter out empty results
        adjusted_qpos_list = [q for q in adjusted_qpos_list if q.numel() > 0]

        return (
            torch.stack(adjusted_qpos_list).to(qpos_list_split.device)
            if adjusted_qpos_list
            else torch.tensor([], device=self.device)
        )

    @ensure_pose_shape
    def get_ik(
        self,
        target_xpos: torch.Tensor,
        qpos_seed: torch.Tensor | None = None,
        num_samples: int | None = None,
        return_all_solutions: bool = False,
        **kwargs,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        r"""Computes the inverse kinematics for given target poses.

        This function generates random joint configurations within the specified limits,
        including the provided joint_seed, and attempts to find valid inverse kinematics solutions.
        It then identifies the joint positions that are closest to the joint_seed.

        Args:
            target_xpos (torch.Tensor): A tensor representing the target positions. It can be of shape
                                         (batch_size, 3) for multiple positions or (3,) for a single position.
            qpos_seed (torch.Tensor | None): Initial joint positions used as seed for IK solving.
                                            Can be:
                                            - 1D tensor of shape (dof,): Single seed for all target positions
                                            - 2D tensor of shape (batch_size, dof): Individual seed per position
                                            If None, defaults to zero configuration. Defaults to None.
            num_samples (int | None): The number of random samples to generate. Must be positive.
                                     Defaults to None.
            return_all_solutions (bool, optional): If True, returns all valid solutions found.
            **kwargs: Additional arguments for future extensions.

        Returns:
            tuple[list[bool], torch.Tensor]: A tuple containing:
                - A tensor of booleans indicating whether valid solutions were found for each target pose. (Shape: (batch_size,))
                - A tensor of shape (batch_size, 1, dof) containing joint positions for
                  each target pose, or an empty tensor if no valid solutions were found.
        """
        # Convert target_pose to tensor and ensure correct device and dtype
        target_xpos = torch.as_tensor(
            target_xpos, device=self.device, dtype=torch.float32
        )
        if num_samples is not None:
            self._num_samples = num_samples

        # Prepare qpos_seed
        if qpos_seed is None:
            qpos_seed = torch.zeros(self.dof, device=self.device)
        else:
            qpos_seed = torch.as_tensor(qpos_seed, device=self.device)

        # Check qpos_seed dimensions
        if qpos_seed.dim() == 1:
            qpos_seed = qpos_seed.unsqueeze(0)
            qpos_seed_ndim = 1
        elif qpos_seed.dim() == 2:
            qpos_seed_ndim = 2
            if qpos_seed.shape[0] != target_xpos.shape[0]:
                raise ValueError(
                    "Batch size of qpos_seed must match batch size of target_xpos when qpos_seed is a 2D tensor."
                )
        else:
            raise ValueError("`qpos_seed` must be a tensor of shape (n,) or (n, n).")

        # Transform target_xpos by TCP
        tcp_xpos = torch.as_tensor(
            deepcopy(self.tcp_xpos), device=self.device, dtype=torch.float32
        )
        target_xpos = target_xpos @ torch.inverse(tcp_xpos)

        # Get joint limits and ensure shape matches dof
        upper_limits = self.upper_position_limits.float()
        lower_limits = self.lower_position_limits.float()

        batch_size = target_xpos.shape[0]

        sampler = QposSeedSampler(
            num_samples=self._num_samples, dof=self.dof, device=self.device
        )
        random_qpos_seeds = sampler.sample(
            qpos_seed, lower_limits, upper_limits, batch_size
        )
        target_xpos_repeated = sampler.repeat_target_xpos(
            target_xpos, self._num_samples
        )

        # Compute IK solutions for all samples
        res_list, qpos_list = self._compute_inverse_kinematics(
            target_xpos_repeated, random_qpos_seeds
        )

        if not isinstance(res_list, torch.Tensor) or not res_list.any():
            logger.log_warning(
                "Pk: No valid solutions found for the given target poses and joint seeds."
            )
            return torch.zeros(
                batch_size, dtype=torch.bool, device=self.device
            ), torch.zeros((batch_size, self.dof), device=self.device)

        # Split res_list and qpos_list according to self._num_samples
        res_list_split = torch.split(res_list, self._num_samples)
        qpos_list_split = torch.split(qpos_list, self._num_samples)

        # Initialize the final results and the closest joint positions
        final_results = []
        final_qpos = []

        # For each batch, select the closest valid solution to qpos_seed
        for i in range(batch_size):
            target_qpos_seed = qpos_seed[i] if qpos_seed_ndim == 2 else qpos_seed

            if not res_list_split[i].any():
                final_results.append(False)
                final_qpos.append(torch.zeros((1, self.dof), device=self.device))
                continue

            result_qpos_limit = self._qpos_to_limits(
                qpos_list_split[i], target_qpos_seed
            )

            if result_qpos_limit.shape[0] == 0:
                final_results.append(False)
                final_qpos.append(torch.zeros((self.dof), device=self.device))
                continue

            distances = torch.norm(result_qpos_limit - target_qpos_seed, dim=1)
            sorted_indices = torch.argsort(distances)
            # shape: (N, dof)
            sorted_qpos_array = result_qpos_limit[sorted_indices]
            final_qpos.append(sorted_qpos_array)
            final_results.append(True)

        # Pad all batches to the same number of solutions for stacking
        max_solutions = max([q.shape[0] for q in final_qpos]) if final_qpos else 1
        final_qpos_tensor = torch.zeros(
            (batch_size, max_solutions, self.dof), device=self.device
        )
        for i, q in enumerate(final_qpos):
            n = q.shape[0]
            final_qpos_tensor[i, :n, :] = q

        final_results = torch.tensor(
            final_results, dtype=torch.bool, device=self.device
        )

        if return_all_solutions:
            # Return all sorted solutions for each batch (shape: batch_size, max_solutions, dof)
            return final_results, final_qpos_tensor

        # Only return the closest solution for each batch (shape: batch_size, 1, dof)
        # If multiple solutions, take the first (closest)
        final_qpos_tensor = final_qpos_tensor[:, :1, :]
        return final_results, final_qpos_tensor

    def get_all_fk(self, qpos: torch.tensor) -> torch.tensor:
        r"""Get the forward kinematics for all links from root to end link.

        Args:
            qpos (torch.Tensor): The joint positions.

        Returns:
            list: A list of 4x4 homogeneous transformation matrices representing the poses of all links from root to end link.
        """
        qpos = torch.as_tensor(qpos)
        qpos = qpos.to(self.device)

        ret = self.pk_serial_chain.forward_kinematics(qpos, end_only=False)
        link_names = list(ret.keys())

        if self.root_link_name is not None:
            try:
                start_index = link_names.index(self.root_link_name)
            except ValueError:
                raise KeyError(
                    f"Root link name '{self.root_link_name}' not found in the kinematic chain"
                )
        else:
            start_index = 0

        if self.end_link_name is not None:
            try:
                end_index = link_names.index(self.end_link_name) + 1
            except ValueError:
                raise KeyError(
                    f"End link name '{self.end_link_name}' not found in the kinematic chain"
                )
        else:
            end_index = len(link_names)

        poses = []
        for link_name in link_names[start_index:end_index]:
            xpos = ret[link_name]
            if not hasattr(xpos, "get_matrix"):
                raise AttributeError(
                    f"The result for link '{link_name}' must have 'get_matrix' attributes."
                )
            xpos_t = torch.eye(4, device=xpos.get_matrix().device)
            m = xpos.get_matrix()
            xpos_t[:3, 3] = m[:, :3, 3]
            xpos_t[:3, :3] = m[:, :3, :3]
            poses.append(xpos_t)

        return poses



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

import numpy as np
import torch
import warp as wp

from typing import Tuple

from embodichain.utils.utility import inv_transform
from embodichain.utils.warp import (
    trajectory_get_diff_kernel,
    trajectory_interpolate_kernel,
    trajectory_add_origin_kernel,
    get_offset_qpos_kernel,
    pairwise_distances,
    cumsum_distances,
    repeat_first_point,
    interpolate_along_distance,
)
from embodichain.lab.sim.solvers.base_solver import BaseSolver
from embodichain.utils.device_utils import standardize_device_string


def compute_pose_offset_related_to_first(full_pose: torch.Tensor) -> torch.Tensor:
    """Compute pose offset relative to the first pose.

    Args:
        full_pose (torch.Tensor): The full pose tensor of shape (N, 4, 4).

    Returns:
        torch.Tensor: The pose offset tensor of shape (N, 4, 4).
    """
    inv_pose0_np = inv_transform(full_pose[0].to("cpu").numpy())
    inv_pose0 = torch.tensor(inv_pose0_np, device=full_pose.device)
    inv_pose0_repeat = inv_pose0[None, :, :].repeat(full_pose.shape[0], 1, 1)
    return torch.bmm(inv_pose0_repeat, full_pose)


def sort_and_padding_key_frame(
    trajectory: np.ndarray, key_indices: np.ndarray, key_frames_batch: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
    """sort and padding key frames for warping trajectory

    Args:
        trajectory (torch.Tensor): raw trajectory. [n_waypoint, dof] of float.
        key_indices (torch.Tensor): key frame waypoint indices. [n_keyframe,] of int.
        key_frames_batch (torch.Tensor): batch key frames. [n_batch, n_keyframe, dof] of float.

    Returns:
        key_indices_ascending (np.ndarray): padded and sorted key frame indices. [n_keyframe_new,] of int.
        key_frames_ascending (np.ndarray): padded and sorted batch key frames. [n_batch, n_keyframe_new, dof] of float.
    """
    sort_ids = np.argsort(key_indices)
    key_indices_ascending = key_indices[sort_ids]
    key_frames_ascending = key_frames_batch[:, sort_ids, :]
    n_batch = key_frames_batch.shape[0]
    if key_indices_ascending[0] != 0:
        key_indices_ascending = np.hstack([0, key_indices_ascending])
        padding_frame = trajectory[0][None, None, :].repeat(n_batch, axis=0)
        key_frames_ascending = np.concatenate(
            [padding_frame, key_frames_ascending], axis=1
        )
    if key_indices_ascending[-1] != trajectory.shape[0] - 1:
        key_indices_ascending = np.hstack(
            [key_indices_ascending, trajectory.shape[0] - 1]
        )
        padding_frame = trajectory[trajectory.shape[0] - 1][None, None, :].repeat(
            n_batch, axis=0
        )
        key_frames_ascending = np.concatenate(
            [key_frames_ascending, padding_frame], axis=1
        )
    return key_indices_ascending, key_frames_ascending


def warp_trajectory_qpos(
    trajectory: torch.Tensor,
    key_indices: torch.Tensor,
    key_frames_batch: torch.Tensor,
    device: str = "cuda",
) -> torch.Tensor:
    """warp trajectory

    Args:
        trajectory (torch.Tensor): raw trajectory. [n_waypoint, dof] of float.
        key_indices (torch.Tensor): key frame waypoint indices. [n_keyframe,] of int.
        key_frames_batch (torch.Tensor): batch key frames. [n_batch, n_keyframe, dof] of float.
        device (str, optional): torch tensor device. Defaults to "cuda".

    Returns:
        torch.Tensor: warped trajectory. [n_batch, n_waypoint, dof] of float.
    """
    # sort and pad key frames
    trajectory_np = trajectory.to("cpu").numpy().astype(np.float32)
    key_indices_np = key_indices.to("cpu").numpy().astype(np.int32)
    key_frames_batch_np = key_frames_batch.to("cpu").numpy().astype(np.float32)

    key_indices_padded, key_frames_padded = sort_and_padding_key_frame(
        trajectory_np, key_indices_np, key_frames_batch_np
    )

    # allocate cuda memory
    n_batch = key_frames_padded.shape[0]
    n_keyframe = key_indices_padded.shape[0]
    n_waypoint, dof = trajectory_np.shape
    wp_in_trajectory = wp.array(
        trajectory_np.flatten(), dtype=float, device=standardize_device_string(device)
    )
    out_trajectory = np.zeros((n_batch, n_waypoint, dof), dtype=np.float32)
    wp_out_trajectory = wp.array(
        out_trajectory.flatten(), dtype=float, device=standardize_device_string(device)
    )
    wp_key_indices = wp.array(
        key_indices_padded, dtype=int, device=standardize_device_string(device)
    )
    wp_key_frames = wp.array(
        key_frames_padded.flatten(),
        dtype=float,
        device=standardize_device_string(device),
    )

    # calcuate
    wp.launch(
        kernel=trajectory_get_diff_kernel,
        dim=(n_batch, dof),
        inputs=[
            wp_in_trajectory,
            wp_key_indices,
            wp_key_frames,
            n_waypoint,
            dof,
            n_keyframe,
        ],
        outputs=[
            wp_out_trajectory,
        ],
        device=standardize_device_string(device),
    )
    wp.launch(
        kernel=trajectory_interpolate_kernel,
        dim=(n_batch, n_waypoint, dof),
        inputs=[wp_key_indices, n_waypoint, dof, n_keyframe],
        outputs=[
            wp_out_trajectory,
        ],
        device=standardize_device_string(device),
    )
    wp.launch(
        kernel=trajectory_add_origin_kernel,
        dim=(n_batch, n_waypoint, dof),
        inputs=[wp_in_trajectory, n_waypoint, dof],
        outputs=[
            wp_out_trajectory,
        ],
        device=standardize_device_string(device),
    )
    warp_traj = (
        wp.to_torch(wp_out_trajectory)
        .reshape(n_batch, n_waypoint, dof)
        .to(torch.device(device))
    )
    return warp_traj


def get_trajectory_object_offset_qpos(
    trajectory: torch.Tensor,
    key_indices: torch.Tensor,
    key_obj_indices: torch.Tensor,
    obj_offset: torch.Tensor,
    solver: BaseSolver,
    base_xpos: torch.Tensor,
    device=torch.device("cuda"),
):
    """warp trajectory according to object pose offset

    Args:
        trajectory (torch.Tensor): raw trajectory. [n_waypoint, dof] of float, joint positions.
        key_indices (torch.Tensor): key frame waypoint indices. [n_keyframe,] of int.
        key_obj_indices (torch.Tensor): key frame belong to which object index. [n_keyframe,] of int.
        obj_offset (torch.Tensor): each object pose offset. [obj_num, n_batch, 4, 4] of float.
        solver (BaseSolver): robot kinematic solver.
        base_xpos (torch.Tensor): solver root link pose in world coordinate. [4, 4] of float.
        device (str, optional): torch tensor device. Defaults to "cuda".

    Returns:
        torch.Tensor: warped trajectory. [n_batch, n_waypoint, dof] of float.
    """
    assert key_indices.shape[0] == key_obj_indices.shape[0]
    dof = trajectory.shape[1]
    key_qpos = trajectory[key_indices]  # [n_keyframe, DOF]
    n_batch = obj_offset.shape[1]  # batch num, aws arena num
    n_keyframe = key_qpos.shape[0]
    key_xpos = solver.get_fk(key_qpos)  # [n_keyframe, 4, 4]

    base_xpos_repeat = base_xpos[None, :, :].repeat(n_keyframe, 1, 1)
    key_xpos = torch.bmm(base_xpos_repeat, key_xpos)

    base_xpos_inv_np = inv_transform(base_xpos.to("cpu").numpy())
    base_xpos_inv_wp = wp.mat44f(base_xpos_inv_np)
    key_obj_indices_wp = wp.from_torch(key_obj_indices.reshape(-1))
    obj_offset_wp = wp.from_torch(obj_offset.reshape(-1))
    key_xpos_wp = wp.from_torch(key_xpos.reshape(-1))
    key_obj_offset_wp = wp.zeros(
        n_batch * n_keyframe * 16, dtype=float, device=standardize_device_string(device)
    )

    wp.launch(
        kernel=get_offset_qpos_kernel,
        dim=(n_batch, n_keyframe),
        inputs=[
            key_obj_indices_wp,
            obj_offset_wp,
            key_xpos_wp,
            base_xpos_inv_wp,
            n_batch,
            n_keyframe,
        ],
        outputs=[
            key_obj_offset_wp,
        ],
        device=standardize_device_string(device),
    )
    key_xpos_offset = wp.to_torch(key_obj_offset_wp).reshape(n_batch * n_keyframe, 4, 4)
    key_qpos_batch = key_qpos[None, :, :].repeat(n_batch, 1, 1).reshape(-1, dof)
    # for pytorch solver, ik use qpos seed but not joint seed
    is_success, key_qpos_offset = solver.get_ik(
        target_xpos=key_xpos_offset,
        qpos_seed=key_qpos_batch,
    )
    key_qpos_offset = key_qpos_offset.reshape(n_batch, n_keyframe, -1)
    return is_success, key_qpos_offset


def interpolate_with_distance(
    trajectory: torch.Tensor,  # expected shape [B, N, M], float or convertible to float
    interp_num: int,  # T
    device=torch.device("cuda"),
) -> torch.Tensor:
    """
    Resample a batch of trajectories of shape [B, N, M] into [B, T, M] by
    piecewise-linear interpolation over cumulative Euclidean distance
    along the N dimension, handling each batch independently.

    Args:
        trajectory: Torch.Tensor of shape [B, N, M].
        interp_num: Target number of samples T.
        device: Torch device string ('cpu', 'cuda', 'cuda:0', ...).
        dtype: Working dtype (wp.float32 or wp.float64). Defaults to wp.float32.

    Returns:
        Torch.Tensor of shape [B, T, M] with interpolated trajectories.
    """
    # Flatten input trajectory for warp kernels (avoid multi-dimensional wp.array bugs)
    trajectory_flat = trajectory.contiguous().to(device).view(-1)
    points = wp.from_torch(trajectory_flat)

    B, N, M = trajectory.shape  # original shape components
    T = int(interp_num)

    if T < 0:
        raise ValueError("`interp_num` must be non-negative.")

    # Handle degenerate T
    out = (
        wp.empty(
            (B * T * M,), dtype=wp.float32, device=standardize_device_string(device)
        )
        if T > 0
        else wp.empty((0,), dtype=wp.float32, device=standardize_device_string(device))
    )

    # Handle N < 2
    if N < 2:
        if N == 1 and T > 0:
            # Repeat the single point across T (kernel expects flattened arrays)
            wp.launch(
                kernel=repeat_first_point,
                dim=B * T,
                inputs=[points, out, B, T, M, N],
                device=standardize_device_string(device),
            )
        # N == 0 -> return empty (out already allocated)
        interp_trajectory = (
            wp.to_torch(out).view(B, T, M) if T > 0 else wp.to_torch(out).view(B, 0, M)
        )
        return interp_trajectory

    if T == 0:
        return out  # nothing to do

    # 1) pairwise distances along N
    dists = wp.empty(
        (B * (N - 1),), dtype=wp.float32, device=standardize_device_string(device)
    )
    wp.launch(
        kernel=pairwise_distances,
        dim=B * (N - 1),
        inputs=[points, dists, B, N, M],
        device=standardize_device_string(device),
    )

    # 2) cumulative distances per batch
    cumulative = wp.empty(
        (B * N,), dtype=wp.float32, device=standardize_device_string(device)
    )
    wp.launch(
        kernel=cumsum_distances,
        dim=B,
        inputs=[dists, cumulative, B, N],
        device=standardize_device_string(device),
    )

    # 3) interpolation per (b, t)
    wp.launch(
        kernel=interpolate_along_distance,
        dim=B * T,
        inputs=[points, cumulative, out, B, N, M, T],
        device=standardize_device_string(device),
    )

    # wp.synchronize_device(device)
    interp_trajectory = wp.to_torch(out).view(B, T, M)
    return interp_trajectory


def interpolate_with_nums(
    trajectory: torch.Tensor,  # expected shape [B, N, M], float or convertible to float
    interp_nums: torch.Tensor,  # expected shape [N - 1], interp_num in each segment
    device=torch.device("cuda"),
) -> torch.Tensor:
    """
    Each entry ``interp_nums[i] = k`` controls segment ``i`` between
    ``trajectory[:, i, :]`` and ``trajectory[:, i + 1, :]``. For that segment,
    ``k`` samples are generated with interpolation factors
    ``alpha = 0, 1/k, 2/k, ..., (k-1)/k`` (i.e., including the segment start
    and excluding the segment end). The final endpoint
    ``trajectory[:, -1, :]`` is appended once at the end of the result, so
    intermediate segment endpoints are not duplicated.

    Args:
        trajectory: Torch.Tensor of shape [B, N, M].
        interp_nums: Torch.Tensor of shape [N - 1] specifying the number of
            samples per segment, including each segment start and excluding
            its end. Values must be non-negative; a value of 0 means that
            no samples are drawn from that segment (other than the final
            overall endpoint that is always appended once).
        device: Torch device string ('cpu', 'cuda', 'cuda:0', ...).

    Returns:
        Torch.Tensor of interpolated trajectories.
    """
    trajectory = trajectory.to(device)
    if not torch.is_floating_point(trajectory):
        trajectory = trajectory.float()

    B, N, M = trajectory.shape
    if N == 0:
        return trajectory.new_empty((B, 0, M))

    interp_nums_tensor = torch.as_tensor(interp_nums, device="cpu").reshape(-1)
    if interp_nums_tensor.numel() != max(N - 1, 0):
        raise ValueError("`interp_nums` must have shape (N - 1,).")

    if N == 1:
        return trajectory[:, :1, :]

    interp_nums_list = interp_nums_tensor.to(torch.int64).tolist()

    # Always seed the output with the first waypoint so it is never dropped,
    # even when leading segments have zero samples.
    segments = [trajectory[:, :1, :]]
    for i, count in enumerate(interp_nums_list):
        if count < 0:
            raise ValueError("`interp_nums` values must be non-negative.")
        p0 = trajectory[:, i : i + 1, :]
        p1 = trajectory[:, i + 1 : i + 2, :]
        if count == 0:
            # No interpolated samples for this segment, but ensure the endpoint
            # waypoint is still present so zero-sample segments don't remove it.
            segments.append(p1)
            continue
        # Generate linearly spaced interpolation parameters from 0 to 1
        # (inclusive), then drop the first value (t = 0) because p0 is
        # already the last point in `segments`. This appends exactly
        # `count` new points per segment and preserves all endpoints.
        alpha = torch.linspace(
            0.0,
            1.0,
            steps=count + 1,
            device=device,
            dtype=trajectory.dtype,
        ).view(1, count + 1, 1)
        seg = p0 + (p1 - p0) * alpha
        segments.append(seg[:, 1:, :])
    return torch.cat(segments, dim=1)
